#
# SchoolTool - common information systems platform for school administration
# Copyright (c) 2014 Shuttleworth Foundation
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"""
Fee views
"""

from decimal import Decimal
from decimal import InvalidOperation
import datetime

import transaction
from zope.browserpage.viewpagetemplatefile import ViewPageTemplateFile
from zope.cachedescriptors.property import Lazy
from zope.component import getUtility
from zope.component import getMultiAdapter
from zope.container.interfaces import INameChooser
from zope.i18n import translate
from zope.i18n.interfaces.locales import ICollator
from zope.security import checkPermission
from zope.security.proxy import removeSecurityProxy
from zope.traversing.browser.absoluteurl import absoluteURL

from z3c.form import button
from z3c.form import field
from z3c.form import form
from z3c.form import widget
from zc.table.column import Column
from zc.table.column import GetterColumn
import zc.resourcelibrary

from schooltool import table
from schooltool.app.states import ACTIVE
from schooltool.app.browser.app import ActiveSchoolYearContentMixin
from schooltool.app.interfaces import ISchoolToolApplication
from schooltool.basicperson.browser.person import BasicPersonTable
from schooltool.basicperson.interfaces import IDemographics
from schooltool.basicperson.interfaces import IDemographicsFields
from schooltool.common import SchoolToolMessage as STMessage
from schooltool.common import getResourceURL
from schooltool.common.inlinept import InheritTemplate
from schooltool.course.interfaces import ISectionContainer
from schooltool.export.importer import FlourishRemoteMegaImporter
from schooltool.export.importer import ERROR_INVALID_PERSON_ID
from schooltool.export.importer import ImportTask
from schooltool.export.importer import ImporterBase
from schooltool.export.importer import RemoteMegaImporter
from schooltool.export.export import RequestXLSReportDialog
from schooltool.export.export import ExcelExportView
from schooltool.group.interfaces import IGroupContainer
from schooltool.person.interfaces import IPerson
from schooltool.report.browser.report import RequestRemoteReportDialog
from schooltool.report.report import ReportLinkViewlet
from schooltool.securitypolicy.crowds import inCrowd
from schooltool.schoolyear.interfaces import ISchoolYear
from schooltool.schoolyear.interfaces import ISchoolYearContainer
from schooltool.schoolyear.subscriber import ObjectEventAdapterSubscriber
from schooltool.stream.interfaces import IGroupConvertedToStreamEvent
from schooltool.stream.interfaces import IStream
from schooltool.stream.interfaces import IStreamContainer
from schooltool.skin import flourish
from schooltool.term.interfaces import IDateManager

from schooltool.fee.fee import Credit
from schooltool.fee.fee import Debit
from schooltool.fee.interfaces import IAmount
from schooltool.fee.interfaces import ICredit
from schooltool.fee.interfaces import IDebit
from schooltool.fee.interfaces import IFeeContainer
from schooltool.fee.interfaces import IReceiptNumbers
# XXX: remove this once the ARK changes get into core
from schooltool.fee.browser.section import is_student

from schooltool.fee import FeeMessage as _


def format_decimal(amount):
    return '%.2f' % amount


class BalanceRow(object):

    amount = 0
    title = _('Balance')
    description = ''

    def __init__(self):
        self.date = datetime.date.max

    @property
    def balance_amount(self):
        return self.amount


class FeePersonBase(object):

    @Lazy
    def person(self):
        return IPerson(self.context)


class PersonFeesViewlet(flourish.viewlet.Viewlet,
                        ActiveSchoolYearContentMixin):

    template = ViewPageTemplateFile('templates/personfees.pt')

    @Lazy
    def future_fees(self):
        result = []
        today = getUtility(IDateManager).today
        for fee in self.fees.values():
            if fee.date > today:
                result.append(fee)
        return result

    @Lazy
    def past_fees(self):
        result = []
        if self.previousYear is not None:
            active = self.active_schoolyear
            active_fees = IFeeContainer(active)
            for fee in self.fees.values():                
                if (fee.__name__ not in active_fees and
                    fee.date < self.active_schoolyear.first):
                    result.append(fee)
        return result

    @property
    def canModify(self):
        return flourish.canEdit(IFeeContainer(self.context))

    @Lazy
    def fees(self):
        return IFeeContainer(self.context)

    def render(self, *args, **kw):
        if is_student(self.context):
            zc.resourcelibrary.need('schooltool.fee.flourish')
            return self.template(*args, **kw)
        return ''


class AmountAddView(flourish.form.AddForm, FeePersonBase):

    template = InheritTemplate(flourish.page.Page.template)
    label = None

    @property
    def title(self):
        return self.person.title

    def updateActions(self):
        super(AmountAddView, self).updateActions()
        self.actions['add'].addClass('button-ok')
        self.actions['cancel'].addClass('button-cancel')

    @button.buttonAndHandler(STMessage('Submit'), name='add')
    def handleAdd(self, action):
        super(AmountAddView, self).handleAdd.func(self, action)

    @button.buttonAndHandler(STMessage('Cancel'))
    def handleCancel(self, action):
        self.request.response.redirect(self.nextURL())

    def nextURL(self):
        return '%s?active_accordion=fees' % absoluteURL(self.person, self.request)

    def create(self, data):
        obj = self.factory()
        form.applyChanges(self, obj, data)
        return obj

    def add(self, obj):
        chooser = INameChooser(self.context)
        name = chooser.chooseName(u'', obj)
        self.context[name] = obj


AddAmountDefaultDate = widget.ComputedWidgetAttribute(
    lambda adapter: adapter.request.util.today,
    view=AmountAddView,
    field=IAmount['date']
    )


AddAmountDefaultAmount = widget.StaticWidgetAttribute(
    0,
    view=AmountAddView,
    field=IAmount['amount']
    )


field_order = ['date', 'title', 'description', 'amount']

credit_field_order = ['date', 'receipt', 'title', 'description', 'amount']


class DebitAddView(flourish.page.Page):

    content_template = flourish.templates.File('templates/debit_add.pt')

    subtitle = _('Add Debit')
    prefix = 'add-debit'

    @property
    def title(self):
        return IPerson(self.context).title

    @property
    def schoolyear(self):
        schoolyears = ISchoolYearContainer(ISchoolToolApplication(None))
        return schoolyears.getActiveSchoolYear()

    @Lazy
    def schoolyear_fees(self):
        result = []
        for fee in sorted(IFeeContainer(self.schoolyear).values(),
                          key=lambda f: f.title):
            result.append({
                'title': fee.title,
                'checked': fee.__name__ in self.context,
                'name': fee.__name__,
                'input_name': '%s_value' % fee.__name__,
                'input_value': fee.amount,
            })
        return result

    def update(self):
        zc.resourcelibrary.need('schooltool.fee.flourish')
        if 'CANCEL' in self.request:
            url = absoluteURL(IPerson(self.context), self.request)
            self.request.response.redirect(url)
            return
        if 'SUBMIT' in self.request:
            submitted_fees = list(self.request.get(self.prefix, []))
            for fee in IFeeContainer(self.schoolyear).values():
                if fee.__name__ in submitted_fees:
                    if fee.__name__ not in self.context:
                        amount = convert_value(self.request.get('%s_value' % fee.__name__))
                        if amount is None:
                            url = '%s/add_debit.html' % absoluteURL(self.context, self.request)
                            transaction.abort()
                            self.request.response.redirect(url + '?ERROR=1')
                            return
                        debit = Debit()
                        debit.amount = amount
                        debit.date = fee.date
                        debit.title = fee.title
                        self.context[fee.__name__] = debit
                elif fee.__name__ in self.context:
                        del self.context[fee.__name__]
            url = '%s?active_accordion=fees' % absoluteURL(IPerson(self.context),
                                                        self.request)
            self.request.response.redirect(url)


class CreditAddView(AmountAddView):

    legend = _('Credit Information')
    fields = field.Fields(ICredit).select(*credit_field_order)
    factory = Credit


DefaultReceiptNumber = widget.ComputedWidgetAttribute(
    lambda adapter: unicode(IReceiptNumbers(ISchoolToolApplication(None))),
    view=CreditAddView,
    field=ICredit['receipt']
    )


class AmountEditView(flourish.form.Form, form.EditForm, FeePersonBase):

    template = InheritTemplate(flourish.page.Page.template)
    content_template = ViewPageTemplateFile('templates/amount_form.pt')

    @property
    def title(self):
        return self.person.title

    def update(self):
        form.EditForm.update(self)

    def updateActions(self):
        super(AmountEditView, self).updateActions()
        self.actions['submit'].addClass('button-ok')
        self.actions['cancel'].addClass('button-cancel')

    def nextURL(self):
        try:
            return self.request['camefrom']
        except (KeyError,):
            return '%s?active_accordion=fees' % absoluteURL(self.person,
                                                         self.request)

    @button.buttonAndHandler(STMessage('Submit'), name='submit')
    def handle_submit(self, action):
        super(AmountEditView, self).handleApply.func(self, action)
        if (self.status == self.successMessage or
           self.status == self.noChangesMessage):
            self.request.response.redirect(self.nextURL())

    @button.buttonAndHandler(STMessage('Cancel'), name='cancel')
    def handle_cancel(self, action):
        self.request.response.redirect(self.nextURL())


class DebitEditView(AmountEditView):

    legend = _('Debit Information')
    fields = field.Fields(IDebit).select(
        'date',
        'last_date',
        'title',
        'label',
        'description',
        'amount',
    )


class CreditEditView(AmountEditView):

    legend = _('Credit Information')
    fields = field.Fields(ICredit).select(*credit_field_order)


class AmountActionsLinks(flourish.page.RefineLinksViewlet):

    pass


class AmountDeleteLink(flourish.page.LinkViewlet):

    @property
    def url(self):
        return '%s/delete.html?delete.%s&CONFIRM' % (
            absoluteURL(self.context.__parent__, self.request),
            self.context.__name__.encode('utf-8'))


class FeeContainerDeleteView(flourish.containers.ContainerDeleteView,
                             FeePersonBase):

    def nextURL(self):
        if 'CONFIRM' in self.request:
            return '%s?active_accordion=fees' % absoluteURL(self.person,
                                                         self.request)
        return flourish.containers.ContainerDeleteView.nextURL(self)


def title_cell_formatter(value, item, formatter):
    if isinstance(item, BalanceRow):
        return value
    return table.table.url_cell_formatter(value, item, formatter)


def amount_cell_formatter(value, item, formatter):
    return format_decimal(value)


def date_cell_formatter(value, item, formatter):
    if isinstance(item, BalanceRow):
        return ''
    return table.table.url_cell_formatter(value, item, formatter)


def edit_cell_formatter(value, item, formatter):
    if isinstance(item, BalanceRow):
        return ''
    template = '''
      <a class="modify" href="%(edit_url)s" title="%(link_title)s">
        <img src="%(edit_icon_url)s" alt="%(edit_title)s" />
      </a>'''
    request = formatter.request
    library = 'schooltool.skin.flourish'
    item_url = absoluteURL(item, request)
    person_url = '%s?active_accordion=fees' % absoluteURL(
        IPerson(item), request)
    return template % {
        'edit_url': '%s/edit.html?camefrom=%s' % (item_url, person_url),
        'link_title': translate(_('Edit this amount'), context=request),
        'edit_icon_url': getResourceURL(library, 'edit-icon.png', request),
        'edit_title': translate(STMessage('Edit'), context=request),
    }


class FeeDateColumn(table.column.DateColumn):

    pass


class FeeAmountColumn(GetterColumn):

    pass


class FeeContainerTable(table.ajax.Table):

    minimum_rows = 5
    batch_size = 0

    @property
    def visible_column_names(self):
        result = ['action', 'date', 'title', 'amount']
        if not self.canModify:
            result = result[1:]
        return result

    def sortOn(self):
        return (('date', False), ('fee_type', False))

    def columns(self):
        actions = self.getActionColumns()
        date = FeeDateColumn(
            name='date',
            title=STMessage('Date'),
            getter=lambda i, f: i.date,
            cell_formatter=date_cell_formatter,
            subsort=True)
        fee_type = GetterColumn(
            name='fee_type',
            title=u'Fee type',
            getter=lambda i, f: ICredit.providedBy(i),
            subsort=True)
        title = GetterColumn(
            name='title',
            title=STMessage('Title'),
            getter=lambda i, f: i.title or '',
            cell_formatter=title_cell_formatter)
        amount = FeeAmountColumn(
            name='amount',
            title=_('Amount'),
            getter=lambda i, f: i.balance_amount,
            cell_formatter=amount_cell_formatter)
        return actions + [date, fee_type, title, amount]

    @Lazy
    def is_manager(self):
        return inCrowd(self.request.principal, 'managers', self.context)

    @Lazy
    def canModify(self):
        return self.is_manager

    def getActionColumns(self):
        if self.canModify:
            return [
                GetterColumn(
                    name='action',
                    title=STMessage('Edit'),
                    getter=lambda i, f: i,
                    cell_formatter=edit_cell_formatter)
            ]
        return []

    def updateFormatter(self):
        if self._table_formatter is None:
            self.setUp(table_formatter=self.table_formatter,
                       batch_size=self.batch_size,
                       prefix=self.__name__,
                       css_classes={'table': 'data fees'})


class FeeContainerTableFilter(table.ajax.TableFilter):

    def filter(self, items):
        if items:
            items = [item for item in items
                     if item not in self.view.future_fees and
                     item not in self.view.past_fees]
            balance = BalanceRow()
            balance.amount = sum([item.balance_amount for item in items])
            items.append(balance)
        return items

    def render(self, *args, **kw):
        return ''


class FutureFeeContainerTable(FeeContainerTable):

    pass


class PastFeeContainerTable(FeeContainerTable):

    pass


class FutureFeeContainerTableFilter(table.ajax.TableFilter):

    def filter(self, items):
        if items:
            items = [item for item in items
                     if item in self.view.future_fees]
            balance = BalanceRow()
            balance.amount = sum([item.balance_amount for item in items])
            items.append(balance)
        return items

    def render(self, *args, **kw):
        return ''


class PastFeeContainerTableFilter(table.ajax.TableFilter):

    def filter(self, items):
        if items:
            items = [item for item in items
                     if item in self.view.past_fees]
            balance = BalanceRow()
            balance.amount = sum([item.balance_amount for item in items])
            items.append(balance)
        return items

    def render(self, *args, **kw):
        return ''


class AmountFeesTableFilter(FeeContainerTableFilter):

    def filter(self, items):
        if items:
            items = list(items)
            balance = BalanceRow()
            balance.title = STMessage('Total')
            balance.amount = sum([item.amount for item in items])
            items.append(balance)
        return items


class AmountView(flourish.form.DisplayForm, FeePersonBase):

    template = InheritTemplate(flourish.page.Page.template)
    content_template = ViewPageTemplateFile('templates/amount.pt')

    @property
    def fields(self):
        if ICredit.providedBy(self.context):
            return field.Fields(ICredit).select(*credit_field_order).omit('title')
        return field.Fields(IAmount).select(*field_order).omit('title')

    @property
    def canModify(self):
        return checkPermission('schooltool.edit', self.context)

    @property
    def title(self):
        return self.person.title

    def nextURL(self):
        return '%s?active_accordion=fees' % absoluteURL(self.person, self.request)

    def edit_url(self):
        item_url = absoluteURL(self.context, self.request)
        return '%(url)s/edit.html?camefrom=%(url)s' % {'url': item_url}


class AmountDoneLink(flourish.viewlet.Viewlet):

    template = flourish.templates.Inline('''
    <h3 class="done-link" i18n:domain="schooltool">
      <a tal:attributes="href view/view/nextURL"
         i18n:translate="">Done</a>
    </h3>
    ''')


class RequestAmountView(RequestRemoteReportDialog):

    report_builder = 'amount.pdf'


class RequestPersonFeeSummaryView(RequestRemoteReportDialog):

    report_builder = 'person_fee_summary.pdf'


class PersonFeeSummaryPDFView(flourish.report.PlainPDFPage, FeePersonBase):

    name = _('Fee Summary')
    message_title = _('fee summary report')

    @property
    def canModify(self):
        return checkPermission('schooltool.edit', self.context)

    def formatDate(self, date, format='mediumDate'):
        if date is None:
            return ''
        formatter = getMultiAdapter((date, self.request), name=format)
        return formatter()

    @property
    def scope(self):
        dtm = getUtility(IDateManager)
        today = dtm.today
        return self.formatDate(today)

    @property
    def title(self):
        return self.person.title

    @property
    def base_filename(self):
        return '%s_fee_summary' % (self.context.username)


class ReceiptPDFView(PersonFeeSummaryPDFView):

    @property
    def name(self):
        if IPerson.providedBy(self.person):
            return translate(_('Receipt ${number}',
                               mapping={'number': self.context.receipt or ''}),
                             context=self.request)
        return _('Activity Credit')

    @property
    def message_title(self):
        if IPerson.providedBy(self.person):
            return _('receipt report')
        return _('activity credit report')

    @property
    def base_filename(self):
        return '%s_%s_receipt' % (self.person.__name__,
                                  self.context.date.strftime('%Y-%m-%d'))


class InvoicePDFView(PersonFeeSummaryPDFView):

    @property
    def name(self):
        if IPerson.providedBy(self.person):
            return _('Invoice')
        return _('Activity Debit')

    @property
    def message_title(self):
        if IPerson.providedBy(self.person):
            return _('invoice report')
        return _('activity debit report')

    @property
    def base_filename(self):
        return '%s_%s_invoice' % (self.person.__name__,
                                  self.context.date.strftime('%Y-%m-%d'))


class FeeSummaryPart(table.pdf.RMLTablePart):

    template = flourish.templates.XMLFile('rml/fee_summary_part.pt')
    table_name = "fees_table"

    def getColumnWidths(self, rml_columns):
        return '20%,65%,15%'

    def render(self):
        columns = self.getColumns()
        rml_columns = self.getRMLColumns(columns)

        if not rml_columns:
            return ''

        items = self.getItems()

        widths_string = self.getColumnWidths(rml_columns)
        headers = self.getHeaders(rml_columns)
        tables = []
        rows = []
        current_group = None
        for item in items:
            group = table.pdf.minimal_escape_rml(self.formatter.getSubGroup(item))
            if group != current_group:
                if rows:
                    tables.append({
                        'headers': current_group and [[current_group]] or [],
                        'rows': rows,
                        })
                current_group = group
                rows = []
            rows.append([column.renderCell(item, self.formatter)
                         for column in rml_columns])

        if rows:
            tables.append({
                    'headers': current_group and [[current_group]] or [],
                    'rows': rows,
                    })
        if tables and tables[-1].get('rows'):
            item_rows = len(tables[-1]['rows']) - 1 # balance row is included
            empty_rows_needed = self.table.minimum_rows - item_rows
            if empty_rows_needed > 0:
                for i in range(empty_rows_needed):
                    tables[-1]['rows'].insert(-1, ['']*len(rml_columns))
        rml = self.template(
            headers=headers, content=tables,
            col_widths=widths_string)
        return rml


class FeesTable(FeeContainerTable):

    visible_column_names = ['date', 'title', 'amount']

    @Lazy
    def source(self):
        return IFeeContainer(self.context)

    def getActionColumns(self):
        return []


class AmountFeesTable(FeesTable):

    @Lazy
    def source(self):
        return {self.context.__name__: self.context}

    def columns(self):
        date, fee_type, title, amount = super(AmountFeesTable, self).columns()
        amount.getter = lambda i, f: i.amount
        return [date, fee_type, title, amount]


class FeeRMLDateColumn(table.pdf.RMLDateColumn):

    def renderCell(self, item, formatter):
        if isinstance(item, BalanceRow):
            return ''
        return super(FeeRMLDateColumn, self).renderCell(item, formatter)


class FeeRMLAmountColumn(table.pdf.RMLGetterColumn):

    def renderCell(self, item, formatter):
        value = self.column.getter(item, formatter)
        return self.escape(format_decimal(value))


class FeeSummaryReportLinkViewlet(ReportLinkViewlet):

    def render(self, *args, **kw):
        if is_student(self.context):
            return self.template(*args, **kw)
        return ''


class InvoiceReportLink(ReportLinkViewlet):

    @property
    def title(self):
        if IPerson.providedBy(self.view.person):
            return _('Invoice')
        return _('Activity Debit')


class ReceiptReportLink(ReportLinkViewlet):

    @property
    def title(self):
        if IPerson.providedBy(self.view.person):
            return _('Receipt')
        return _('Activity Credit')


class RequestSchoolFeeSummaryView(RequestRemoteReportDialog):

    report_builder = 'fee_summary.pdf'


class SchoolFeeSummaryPDFView(flourish.report.PlainPDFPage):

    name = _('Fee Summary')
    message_title = _('school fee summary report')

    @property
    def canModify(self):
        return checkPermission('schooltool.edit', self.context)

    def formatDate(self, date, format='mediumDate'):
        if date is None:
            return ''
        formatter = getMultiAdapter((date, self.request), name=format)
        return formatter()

    @property
    def scope(self):
        dtm = getUtility(IDateManager)
        today = dtm.today
        return self.formatDate(today)

    @property
    def title(self):
        return self.name

    @property
    def base_filename(self):
        return 'school_fee_summary'

    @Lazy
    def schoolyear_sections(self):
        result = []
        schoolyears = ISchoolYearContainer(self.context)
        active = schoolyears.getActiveSchoolYear()
        if active is not None:
            for term in active.values():
                result.extend(ISectionContainer(term).values())
            result.append(IGroupContainer(active)['students'])
        return result

    def is_student(self, sections, person):
        for section in sections:
            if person in section.members:
                return True
        return False

    def balance(self, person, date):
        fees = [fee for fee in IFeeContainer(person).values()
                if fee.date <= date]
        return sum([item.balance_amount for item in fees])

    def update(self):
        self.students = []
        self.paid_students = []
        self.not_paid_students = []
        dtm = getUtility(IDateManager)
        today = dtm.today
        for person in self.context['persons'].values():
            if self.is_student(self.schoolyear_sections, person):
                self.students.append(person)
                balance = self.balance(person, today)
                info = {'title': person.title, 'balance': balance}
                if balance >= 0:
                    self.paid_students.append(info)
                else:
                    self.not_paid_students.append(info)
        super(SchoolFeeSummaryPDFView, self).update()


class SchoolFeeSummaryHeaderPart(flourish.report.PDFPart):

    template = flourish.templates.XMLFile(
        'rml/school_fee_summary_header.pt')

    def paid_students(self):
        return '%d/%d' % (len(self.view.paid_students),
                          len(self.view.students))


class SchoolFeeSummaryTablePart(flourish.report.PDFPart):

    template = flourish.templates.XMLFile(
        'rml/school_fee_summary_table.pt')

    col_widths = "80%,20%"

    def update(self):
        collator = ICollator(self.request.locale)
        self.not_paid_students = sorted(self.view.not_paid_students,
                                        key=lambda i:collator.key(i['title']))
        super(SchoolFeeSummaryTablePart, self).update()


class SchoolFeeSummaryReportLink(ReportLinkViewlet,
                                 ActiveSchoolYearContentMixin):

    def render(self, *args, **kw):
        if self.schoolyear is None:
            return ''
        return super(SchoolFeeSummaryReportLink, self).render(*args, **kw)


class SchoolYearReportLink(ReportLinkViewlet,
                           ActiveSchoolYearContentMixin):

    @property
    def report_link(self):
        url = absoluteURL(self.schoolyear, self.request)
        return '%s/%s' % (url, self.link)


class FeeContainerLinkViewlet(flourish.page.LinkViewlet,
                              ActiveSchoolYearContentMixin):

    @property
    def enabled(self):
        return self.schoolyear

    @property
    def url(self):
        container = IFeeContainer(self.schoolyear)
        return absoluteURL(container, self.request)


class SchoolYearFeeContainerView(flourish.page.Page):

    content_template = flourish.templates.Inline('''
      <div>
        <tal:block content="structure context/schooltool:content/ajax/table" />
      </div>
    ''')

    @property
    def title(self):
        return self.context.title


class SchoolYearFeeContainerAddLinks(
        flourish.page.RefineLinksSortedByTitleViewlet):

    pass


def fee_title_formatter(value, item, formatter):
    url = '%s/schoolyear_fee.html' % absoluteURL(item, formatter.request)
    return '<a href="%s">%s</a>' % (url, value)


class SchoolYearFeeContainerTableFilter(table.ajax.TableFilter):

    def is_recurrence(self, item):
        return item.master is not None

    def filter(self, items):
        return [item for item in items
                if not self.is_recurrence(item)]

    def render(self, *args, **kw):
        return ''


class SchoolYearFeeContainerTable(table.ajax.Table):

    no_default_url_cell_formatter = True

    def columns(self):
        title = GetterColumn(
            name='title',
            title=STMessage(u'Title'),
            cell_formatter=fee_title_formatter,
            getter=lambda i, f: i.title)
        amount = GetterColumn(
            name='amount',
            title=_('Standard Amount'),
            getter=lambda i, f: format_decimal(i.amount))
        return [title, amount]


from zope.schema import List
from zope.schema import Object
from zope.schema import TextLine
from zope.schema import Date
from z3c.form.interfaces import IObjectWidget
from z3c.form.object import ObjectSubForm
from z3c.form.object import SubformAdapter
from z3c.form.object import registerFactoryAdapter
from zope.interface import implements
from zope.component import adapts
from zope.interface import Interface


class IDebitRecurrence(Interface):

    date = Date(title=STMessage(u'Date'))

    last_date = Date(title=_(u'Last date'))

    title = TextLine(title=STMessage(u'Title'))

    label = TextLine(
        title=STMessage(u'Label'),
        description=_(u'Limit to 7 characters or less'),
        max_length=7)


class DebitRecurrence(object):

    implements(IDebitRecurrence)

    date = None
    last_date = None
    title = None
    label = None


registerFactoryAdapter(IDebitRecurrence, DebitRecurrence)


class RecurrenceSubForm(ObjectSubForm):

    def setupFields(self):
        self.fields = field.Fields(IDebitRecurrence).select(
            'date',
            'last_date',
            'title',
            'label'
        )


class IAddDebit(IDebit):

    recurrences = List(
        title=_(u'Recurrences'),
        value_type=Object(title=_(u'Recurrence'), schema=IDebitRecurrence),
        required=False)


class SchoolYearFeeAddView(flourish.form.AddForm,
                           ActiveSchoolYearContentMixin):

    template = flourish.templates.Inherit(flourish.page.Page.template)
    label = None
    legend = _('Fee Information')

    fields = field.Fields(IAddDebit).select(
        'date',
        'last_date',
        'title',
        'label',
        'description',
        'amount',
        'recurrences')

    def updateActions(self):
        super(SchoolYearFeeAddView, self).updateActions()
        self.actions['add'].addClass('button-ok')
        self.actions['cancel'].addClass('button-cancel')

    @button.buttonAndHandler(STMessage('Submit'), name='add')
    def handleAdd(self, action):
        super(SchoolYearFeeAddView, self).handleAdd.func(self, action)

    @button.buttonAndHandler(STMessage('Cancel'), name='cancel')
    def handleCancel(self, action):
        url = absoluteURL(self.context, self.request)
        self.request.response.redirect(url)

    def create(self, data):
        debit = Debit()
        self._recurrences = data.pop('recurrences')
        form.applyChanges(self, debit, data)
        return debit

    def add(self, debit):
        chooser = INameChooser(self.context)
        name = chooser.chooseName(u'', debit)
        self.context[name] = debit
        if self._recurrences:
            for recurrence in self._recurrences:
                recurrence_debit = Debit()
                data = {
                    'date': recurrence.date,
                    'last_date': recurrence.last_date,
                    'title': recurrence.title,
                    'label': recurrence.label,
                    'description': debit.description,
                    'amount': debit.amount,
                }
                form.applyChanges(self, recurrence_debit, data)
                name = chooser.chooseName(u'', recurrence_debit)
                self.context[name] = recurrence_debit
                recurrence_debit.master = debit.__name__
        self._debit = debit
        return debit

    def nextURL(self):
        return absoluteURL(self.context, self.request)

    def updateWidgets(self):
        super(SchoolYearFeeAddView, self).updateWidgets()
        self.widgets['date'].label = _(u'Due Date')
        self.widgets['amount'].label = _(u'Standard Amount')


class RecurrenceSubformAdapter(SubformAdapter):

    adapts(Interface,
           flourish.IFlourishLayer,
           Interface,
           SchoolYearFeeAddView,           
           IObjectWidget,
           Interface,
           Interface)

    factory = RecurrenceSubForm


class SchoolYearFeeEditView(flourish.form.Form, form.EditForm):

    template = InheritTemplate(flourish.page.Page.template)
    legend = _('Fee Information')
    fields = field.Fields(IDebit).select(
        'date',
        'last_date',
        'title',
        'label',
        'description',
        'amount'
    )

    @property
    def title(self):
        return self.context.title

    def update(self):
        form.EditForm.update(self)

    def updateWidgets(self):
        super(SchoolYearFeeEditView, self).updateWidgets()
        self.widgets['date'].label = _(u'Due Date')
        self.widgets['amount'].label = _(u'Standard Amount')

    def updateActions(self):
        super(SchoolYearFeeEditView, self).updateActions()
        self.actions['submit'].addClass('button-ok')
        self.actions['cancel'].addClass('button-cancel')

    def nextURL(self):
        # XXX: used by amount view done link
        return absoluteURL(self.context.__parent__, self.request)

    @button.buttonAndHandler(STMessage('Submit'), name='submit')
    def handle_submit(self, action):
        data, errors = self.extractData()
        if errors:
            self.status = self.formErrorsMessage
            return
        changes = self.applyChanges(data)
        if changes:
            self.status = self.successMessage
            app = ISchoolToolApplication(None)
            persons = app['persons']
            schoolyear = self.context.__parent__.__parent__
            groups = IGroupContainer(schoolyear)
            streams = IStreamContainer(schoolyear)
            changed_attrs = []
            for iface, attrs in changes.items():
                changed_attrs.extend(attrs)
            for person in persons.values():
                fees = IFeeContainer(person)
                if self.context.__name__ in fees:
                    fee = fees[self.context.__name__]
                    for name in changed_attrs:
                        setattr(fee, name, data[name])
                for fee in fees.values():
                    if (IDebit.providedBy(fee) and
                        fee.master == self.context.__name__):
                        for name in changed_attrs:
                            setattr(fee, name, data[name])
            for group in groups.values():
                fees = IFeeContainer(group)
                if self.context.__name__ in fees:
                    fee = fees[self.context.__name__]
                    for name in changed_attrs:
                        setattr(fee, name, data[name])
            for stream in streams.values():
                fees = IFeeContainer(stream)
                if self.context.__name__ in fees:
                    fee = fees[self.context.__name__]
                    for name in changed_attrs:
                        setattr(fee, name, data[name])
        else:
            self.status = self.noChangesMessage
        if IAmount in changes and 'amount' in changes[IAmount]:
            container = self.context.__parent__
            for item in container.values():
                if item.master == self.context.__name__:
                    item.amount = self.context.amount
        if (self.status == self.successMessage or
           self.status == self.noChangesMessage):
            url = '%s/schoolyear_fee.html' % absoluteURL(self.context, self.request)
            self.request.response.redirect(url)

    @button.buttonAndHandler(STMessage('Cancel'), name='cancel')
    def handle_cancel(self, action):
        url = '%s/schoolyear_fee.html' % absoluteURL(self.context, self.request)
        self.request.response.redirect(url)


class SchoolYearFeeRecurrenceAddView(flourish.form.AddForm):

    template = flourish.templates.Inherit(flourish.page.Page.template)
    label = None
    legend = _('Fee Information')

    @property
    def title(self):
        return self.context.title

    fields = field.Fields(IDebitRecurrence).select(
        'date',
        'last_date',
        'title',
        'label'
    )

    def updateActions(self):
        super(SchoolYearFeeRecurrenceAddView, self).updateActions()
        self.actions['add'].addClass('button-ok')
        self.actions['cancel'].addClass('button-cancel')

    @button.buttonAndHandler(STMessage('Submit'), name='add')
    def handleAdd(self, action):
        super(SchoolYearFeeRecurrenceAddView, self).handleAdd.func(self, action)

    @button.buttonAndHandler(STMessage('Cancel'), name='cancel')
    def handleCancel(self, action):
        url = '%s/schoolyear_fee.html' % absoluteURL(self.context, self.request)
        self.request.response.redirect(url)

    def create(self, data):
        debit = Debit()
        debit.date = data['date']
        debit.last_date = data['last_date']
        debit.title = data['title']
        debit.label = data['label']
        return debit

    def add(self, debit):
        container = self.context.__parent__
        chooser = INameChooser(container)
        name = chooser.chooseName(u'', debit)
        container[name] = debit
        debit.amount = self.context.amount
        debit.description = self.context.description
        debit.master = self.context.__name__
        self._debit = debit
        schoolyear = container.__parent__
        groups = IGroupContainer(schoolyear)
        for group in groups.values():
            group_fees = IFeeContainer(group)
            if self.context.__name__ in group_fees:
                group_fee = Debit()
                group_fee.date = debit.date
                group_fee.last_date = debit.last_date
                group_fee.title = debit.title
                group_fee.label = debit.label
                group_fee.description = debit.description
                group_fee.amount = debit.amount
                group_fee.master = debit.master
                group_fees[name] = group_fee
                for person in group.members:
                    person_fees = IFeeContainer(person)
                    if name not in person_fees:
                        person_fee = Debit()
                        person_fee.date = debit.date
                        person_fee.last_date = debit.last_date
                        person_fee.title = debit.title
                        person_fee.label = debit.label
                        person_fee.description = debit.description
                        person_fee.amount = debit.amount
                        person_fee.master = debit.master
                        person_fees[name] = person_fee
        streams = IStreamContainer(schoolyear)
        for stream in streams.values():
            stream_fees = IFeeContainer(stream)
            if self.context.__name__ in stream_fees:
                stream_fee = Debit()
                stream_fee.date = debit.date
                stream_fee.last_date = debit.last_date
                stream_fee.title = debit.title
                stream_fee.label = debit.label
                stream_fee.description = debit.description
                stream_fee.amount = debit.amount
                stream_fee.master = debit.master
                stream_fees[name] = stream_fee
                for person in stream.members:
                    person_fees = IFeeContainer(person)
                    if name not in person_fees:
                        person_fee = Debit()
                        person_fee.date = debit.date
                        person_fee.last_date = debit.last_date
                        person_fee.title = debit.title
                        person_fee.label = debit.label
                        person_fee.description = debit.description
                        person_fee.amount = debit.amount
                        person_fee.master = debit.master
                        person_fees[name] = person_fee
        return debit

    def nextURL(self):
        return '%s/schoolyear_fee.html' % absoluteURL(self.context, self.request)

    def updateWidgets(self):
        super(SchoolYearFeeRecurrenceAddView, self).updateWidgets()
        self.widgets['date'].label = _(u'Due Date')


class SchoolYearFeeRecurrenceEditView(flourish.form.Form, form.EditForm):

    template = InheritTemplate(flourish.page.Page.template)
    legend = _('Fee Information')
    fields = field.Fields(IDebit).select(
        'date',
        'last_date',
        'title',
        'label',
    )

    @property
    def title(self):
        return self.context.title

    def update(self):
        form.EditForm.update(self)

    def updateWidgets(self):
        super(SchoolYearFeeRecurrenceEditView, self).updateWidgets()
        self.widgets['date'].label = _(u'Due Date')

    def updateActions(self):
        super(SchoolYearFeeRecurrenceEditView, self).updateActions()
        self.actions['submit'].addClass('button-ok')
        self.actions['cancel'].addClass('button-cancel')

    def nextURL(self):
        url = '%s/%s/schoolyear_fee.html' % (
            absoluteURL(self.context.__parent__, self.request),
            self.context.master)
        return self.request.get('camefrom', url)

    @button.buttonAndHandler(STMessage('Submit'), name='submit')
    def handle_submit(self, action):
        data, errors = self.extractData()
        if errors:
            self.status = self.formErrorsMessage
            return
        changes = self.applyChanges(data)
        if changes:
            self.status = self.successMessage
            app = ISchoolToolApplication(None)
            persons = app['persons']
            schoolyear = self.context.__parent__.__parent__
            groups = IGroupContainer(schoolyear)
            streams = IStreamContainer(schoolyear)
            changed_attrs = []
            for iface, attrs in changes.items():
                changed_attrs.extend(attrs)
            for person in persons.values():
                fees = IFeeContainer(person)
                if self.context.__name__ in fees:
                    fee = fees[self.context.__name__]
                    for name in changed_attrs:
                        setattr(fee, name, data[name])
            for group in groups.values():
                fees = IFeeContainer(group)
                if self.context.__name__ in fees:
                    fee = fees[self.context.__name__]
                    for name in changed_attrs:
                        setattr(fee, name, data[name])
            for stream in streams.values():
                fees = IFeeContainer(stream)
                if self.context.__name__ in fees:
                    fee = fees[self.context.__name__]
                    for name in changed_attrs:
                        setattr(fee, name, data[name])
        else:
            self.status = self.noChangesMessage
        if (self.status == self.successMessage or
           self.status == self.noChangesMessage):
            self.request.response.redirect(self.nextURL())

    @button.buttonAndHandler(STMessage('Cancel'), name='cancel')
    def handle_cancel(self, action):
        self.request.response.redirect(self.nextURL())
                

class GroupFeeView(flourish.page.Page):

    container_class = 'container widecontainer'

    def update(self):
        zc.resourcelibrary.need('schooltool.fee.flourish')

    @property
    def title(self):
        return self.context.title

    @Lazy
    def schoolyear_fees(self):
        schoolyear = ISchoolYear(self.context.__parent__)
        return IFeeContainer(schoolyear)


class GroupAddFeesTableFilter(table.ajax.TableFilter):

    template = flourish.templates.File('templates/group_fee_table_filter.pt')

    def fees(self):
        result = []
        for fee_id, fee in self.view.schoolyear_fees.items():
            if not fee.master and fee.__name__ not in IFeeContainer(self.context):
                result.append({
                    'id': fee_id,
                    'title': fee.title,
                })
        return result

    def filter(self, items):
        if self.manager.selected_fee is None:
            return items
        return [item for item in items
                if not self.has_fee(item, self.manager.selected_fee)]

    def has_fee(self, item, fee):
        person = table.column.unindex(item)
        return fee.__name__ in IFeeContainer(person)


class FeeCheckboxColumn(table.column.CheckboxColumn):

    def template(self):
        return '<input type="checkbox" name="%(name)s" value="%(value)s" id="%(id)s"%(checked)s />'

    def params(self, item, formatter):
        selected_fee = formatter.view.selected_fee
        checked = (selected_fee.__name__ in IFeeContainer(item)
                   if selected_fee is not None else False)
        form_id = ".".join(filter(None, [self.prefix, self.id_getter(item)]))
        return {
            'name': self.prefix + ':list',
            'value': item.__name__,
            'id': form_id,
            'checked': checked and ' checked="checked"' or '',
            }


def input_name(prefix, person_id):
    return prefix + '.' + person_id


class FeeAmountInputColumn(Column):

    def __init__(self, prefix, name=None, title=None):
        super(FeeAmountInputColumn, self).__init__(name=name, title=title)
        self.prefix = prefix

    def template(self):
        return '<input type="text" class="%(klass)s" name="%(name)s" id="%(id)s" value="%(value)s" />'

    def params(self, item, formatter):
        selected_fee = formatter.view.selected_fee
        value = ''
        if selected_fee is not None:
            fee = IFeeContainer(item).get(selected_fee.__name__)
            if fee is not None:
                value = fee.amount
            else:
                value = selected_fee.amount
        name = input_name(self.prefix, item.__name__)
        klass = ['fee-input']
        if item.__name__ in formatter.view.errors:
            klass.append('fee-input-error')
        return {
            'klass': ' '.join(klass),
            'name': name,
            'id': name,
            'value': value
            }

    def renderCell(self, item, formatter):
        params = self.params(item, formatter)
        return self.template() % params


def convert_value(value):
    try:
        return Decimal(value)
    except (InvalidOperation, TypeError):
        return


class GroupAddFeesTable(BasicPersonTable):

    batch_size = 0

    def __init__(self, *args, **kw):
        super(GroupAddFeesTable, self).__init__(*args, **kw)
        self.errors = {}

    @property
    def search_fee_id(self):
        return self.html_id + '-fee'

    @property
    def action_name(self):
        return self.html_id + '-action'

    @property
    def submit_name(self):
        return self.html_id + '-submit'

    @property
    def cancel_name(self):
        return '___CANCEL___'

    @property
    def input_prefix(self):
        return self.html_id + '-input'

    def items(self):
        return self.indexItems(self.context.members)

    @Lazy
    def selected_fee(self):
        requested_fee = self.request.get(self.search_fee_id)
        return self.view.schoolyear_fees.get(requested_fee)

    def nextURL(self):
        return absoluteURL(self.context, self.request)

    def columns(self):
        default = BasicPersonTable.columns(self)
        checkbox = FeeCheckboxColumn(
            self.action_name,
            name='checkbox',
            title=_('Apply?'),
            id_getter=lambda i: i.__name__)
        amount = FeeAmountInputColumn(
            self.input_prefix,
            name='amount',
            title=_('Amount'))
        weakness = GetterColumn(
            name='weakness',
            title=_('EWS / Freeship'),
            getter=lambda i, f: IDemographics(i).get('economic_weakness') or '')
        result = [checkbox, amount] + default
        if self.add_weakness_column:
            result =  result + [weakness]
        return result

    @Lazy
    def add_weakness_column(self):
        app = ISchoolToolApplication(None)
        dfs = IDemographicsFields(app)
        return 'economic_weakness' in dfs

    def update(self):
        super(GroupAddFeesTable, self).update()
        view_url = '%s/add_fee.html' % absoluteURL(self.context, self.request)
        if self.cancel_name in self.request:
            url = absoluteURL(self.context, self.request)
            self.request.response.redirect(url)
            return
        if self.submit_name in self.request:
            if self.selected_fee is not None:
                submitted_persons = list(self.request.get(self.action_name, []))
                errors = self.validate(submitted_persons)
                if not errors:
                    self.process(submitted_persons)
                    container = IFeeContainer(self.context)
                    if self.selected_fee.__name__ not in container:
                        fee = Debit()
                        fee.date = self.selected_fee.date
                        fee.last_date = self.selected_fee.last_date
                        fee.label = self.selected_fee.label
                        fee.title = self.selected_fee.title
                        fee.description = self.selected_fee.description
                        fee.amount = self.selected_fee.amount
                        container[self.selected_fee.__name__] = fee
                        for recurrence in self.recurrences(self.selected_fee):
                            recurrence_fee = Debit()
                            recurrence_fee.date = recurrence.date
                            recurrence_fee.last_date = recurrence.last_date
                            recurrence_fee.title = recurrence.title
                            recurrence_fee.label = recurrence.label
                            recurrence_fee.description = recurrence.description
                            recurrence_fee.amount = recurrence.amount
                            recurrence_fee.master = recurrence.master
                            container[recurrence.__name__] = recurrence_fee
                    url = absoluteURL(self.context, self.request)
                    self.request.response.redirect(url)
                    return
                else:
                    self.request.response.redirect(view_url + '?ERROR=1')
                    return
            else:
                self.request.response.redirect(view_url)

    def validate(self, submitted_persons):
        for person_id in submitted_persons:
            value_name = input_name(self.input_prefix, person_id)
            value = convert_value(self.request.get(value_name))
            if value is None:
                self.errors[person_id] = None
        return self.errors

    def process(self, submitted_persons):
        for item in self._items:
            person = table.column.unindex(item)
            checked = person.__name__ in submitted_persons
            fee = IFeeContainer(person).get(self.selected_fee.__name__)
            if fee is not None:
                if checked:
                    value_name = input_name(self.input_prefix, person.__name__)
                    value = convert_value(self.request.get(value_name))
                    if value != fee.amount:
                        fee.amount = value
                else:
                    del IFeeContainer(person)[self.selected_fee.__name__]
            else:
                if checked:
                    fee = Debit()
                    value_name = input_name(self.input_prefix, person.__name__)
                    fee.amount = convert_value(self.request.get(value_name))
                    fee.title = self.selected_fee.title
                    fee.date = self.selected_fee.date
                    fee.last_date = self.selected_fee.last_date
                    fee.label = self.selected_fee.label
                    fee.description = self.selected_fee.description
                    IFeeContainer(person)[self.selected_fee.__name__] = fee
                    for recurrence in self.recurrences(self.selected_fee):
                        recurrence_fee = Debit()
                        recurrence_fee.date = recurrence.date
                        recurrence_fee.last_date = recurrence.last_date
                        recurrence_fee.title = recurrence.title
                        recurrence_fee.label = recurrence.label
                        recurrence_fee.description = recurrence.description
                        recurrence_fee.amount = fee.amount
                        recurrence_fee.master = recurrence.master
                        name = recurrence.__name__
                        IFeeContainer(person)[name] = recurrence_fee

    def recurrences(self, fee):
        result = []
        container = fee.__parent__
        for item in container.values():
            if item.master == fee.__name__:
                result.append(item)
        return result


class GroupFeeTableButtons(flourish.viewlet.Viewlet):

    template = flourish.templates.File('templates/group_fee_table_buttons.pt')

    submit_title = STMessage('Submit')
    cancel_title = STMessage('Cancel')

    @property
    def select_all_name(self):
        return self.manager.html_id + '-select-all'

    @property
    def select_none_name(self):
        return self.manager.html_id + '-select-none'


class AddFeeLinkViewlet(flourish.page.LinkViewlet):

    @property
    def enabled(self):
        schoolyear = ISchoolYear(self.context.__parent__)
        return IFeeContainer(schoolyear)


from zope.interface import Interface
from zope.interface import implements
from zope.schema import Choice
from zope.schema.interfaces import IContextSourceBinder
from zope.schema.vocabulary import SimpleVocabulary
from schooltool.group.group import defaultGroups
import xlwt
from StringIO import StringIO


class IBasicFeesReport(Interface):

    group = Choice(
        title=STMessage('Stream'),
        source='schooltool.stream.streams')


class BasicFeesReportRequestView(RequestXLSReportDialog):

    report_builder = 'basic_fees_report.xls'

    fields = field.Fields(IBasicFeesReport)

    def resetForm(self):
        RequestXLSReportDialog.resetForm(self)
        self.form_params['schoolyear'] = self.context

    def updateTaskParams(self, task):
        group= self.form_params.get('group')
        task.request_params['group'] = group.__name__


class BasicFeesReportView(ExcelExportView):

    message_title = _('Fees report: Basic fees report by section')

    sheet_1_title = _('Basic fees report by section with comparison across classes')
    sheet_1_label = _('Fees module')

    overall_line_id = 'overall'
    _style_cache = {}

    def easyxf(self, s, num_format_str=None):
        key = s.strip(), num_format_str
        if key not in self._style_cache:
            self._style_cache[key] = xlwt.easyxf(s, num_format_str)
        return self._style_cache[key]

    def addExporters(self, progress):
        progress.add(
            'overall', title=_('Report'), progress=0.0)
        progress.add(
            'data', title=_('Gathering report data'),
            active=False, progress=0.0)

    @property
    def base_filename(self):
        return 'basic_fees_%s' % self.schoolyear.title

    @Lazy
    def streams(self):
        streams = IStreamContainer(self.schoolyear)
        return sorted(streams.values(), key=lambda g: g.title)

    def group_data(self, group):
        result = []
        today = getUtility(IDateManager).today
        for person in group.members:
            fees = {}
            person_fees = IFeeContainer(person)
            for fee in self.schoolyear_fees.values():
                if fee.__name__ in person_fees and fee.date <= today:
                    fees[fee.__name__] = person_fees[fee.__name__]
            total_debit = self.person_total_debit(
                fees, self.schoolyear.first, today)
            total_credit = self.person_total_credit(
                person, self.schoolyear.first, today)
            info = {
                'gender': person.gender,
                'title': person.title,
                'fees': fees,
                'total_debit': total_debit,
                'total_credit': total_credit,
                'balance': total_credit - total_debit,
                'previous': 0,
            }
            result.append(info)
        return result

    def export_report(self, wb):
        self.task_progress.force('sheet1', active=True)
        ws = wb.add_sheet(self.sheet_1_label)
        self.print_headers(ws, self.sheet_1_title)
        self.print_legend_1(ws, row=2, col=self.merge - 2)
        row = 7
        title = _('Comparison across classes / sections')
        ws.write_merge(row, row, 0, self.merge, title, self.easyxf(
            'font: bold on, color white; '
            'pattern: pattern solid, fore_color report_title_bg; '
            'borders: top no_line, right no_line, bottom no_line, left no_line; '
        ))
        row += 1
        title = _('Fees owed')
        ws.write_merge(row, row, 1, self.merge - 4, title, self.easyxf(
            'font: bold on, color white; '
            'pattern: pattern solid, fore_color owed_bg; '
            'alignment: horizontal center; '
            'borders: top no_line, right no_line, bottom thin, left no_line; '
        ))
        ws.write(row, self.merge - 3, _('Fees paid'), self.easyxf(
            'font: bold on, color white; '
            'pattern: pattern solid, fore_color paid_bg; '
            'alignment: horizontal center; '
            'borders: top no_line, right no_line, bottom thin, left no_line; '
        ))
        title = _('Outstanding balance')
        ws.write_merge(row, row, self.merge - 2, self.merge, title, self.easyxf(
            'font: bold on, color white; '
            'pattern: pattern solid, fore_color balance_bg; '
            'alignment: horizontal center; '
            'borders: top no_line, right no_line, bottom thin, left no_line; '
        ))
        row += 1
        ws.write(row, 0, '', self.easyxf(
            'font: bold on, color black; '
            'pattern: pattern solid, fore_color white; '
            'borders: top no_line, right no_line, bottom thin, left no_line; '
        ))
        col = 1
        for fee in self.collected_fees:
            ws.write(row, col, '', self.easyxf(
                'font: bold on, color black; '
                'pattern: pattern solid, fore_color white; '
                'borders: top no_line, right no_line, bottom thin, left no_line; '
            ))
            col += 1
        ws.write(row, col, _('Total YTD*'), self.easyxf(
            'font: bold on, color white; '
            'pattern: pattern solid, fore_color owed_bg; '
            'alignment: horizontal center; '
            'borders: top no_line, right no_line, bottom thin, left no_line; '
        ))
        col += 1
        ws.write(row, col, _('Total YTD*'), self.easyxf(
            'font: bold on, color white; '
            'pattern: pattern solid, fore_color paid_bg; '
            'alignment: horizontal center; '
            'borders: top no_line, right no_line, bottom thin, left no_line; '
        ))
        col += 1
        ws.write(row, col, _('Bal - YTD*'), self.easyxf(
            'font: bold on, color white; '
            'pattern: pattern solid, fore_color balance_bg; '
            'alignment: horizontal center; '
            'borders: top no_line, right no_line, bottom thin, left no_line; '
        ))
        col += 1
        ws.write(row, col, _('% paid'), self.easyxf(
            'font: bold on, color white; '
            'pattern: pattern solid, fore_color balance_bg; '
            'alignment: horizontal center; '
            'borders: top no_line, right no_line, bottom thin, left no_line; '
        ))
        col += 1
        ws.write(row, col, _('Prev. years'), self.easyxf(
            'font: bold on, color white; '
            'pattern: pattern solid, fore_color balance_bg; '
            'alignment: horizontal center; '
            'borders: top no_line, right no_line, bottom thin, left no_line; '
        ))
        row += 1
        school_debit = []
        school_debit_girls = []
        school_debit_boys = []
        school_credit = []
        school_credit_girls = []
        school_credit_boys = []
        school_balance = []
        school_balance_girls = []
        school_balance_boys = []
        school_previous = []
        school_previous_girls = []
        school_previous_boys = []
        for stream in self.streams:
            group_data = self.group_data(stream)
            col = 0
            ws.write(row, col, stream.title, self.easyxf(
                'font: bold on, color black; '
                'pattern: pattern solid, fore_color white; '
                'borders: top no_line, right no_line, bottom no_line, left no_line; '
            ))
            col += 1
            for fee in self.collected_fees:
                ws.write(row, col, '', self.easyxf(
                    'font: bold off, color black; '
                    'pattern: pattern solid, fore_color white; '
                    'borders: top no_line, right no_line, bottom no_line, left no_line; '))
                col += 1
            value = sum([person_info['total_debit'] for person_info in group_data])
            school_debit.append(value)
            school_debit_girls.append(sum([person_info['total_debit'] for person_info in group_data if person_info['gender'] == 'female']))
            school_debit_boys.append(sum([person_info['total_debit'] for person_info in group_data if person_info['gender'] == 'male']))
            ws.write(row, col, value or '', self.easyxf(
                'font: bold on, color black; '
                'pattern: pattern solid, fore_color white; '
                'borders: top no_line, right no_line, bottom no_line, left no_line; '
            ))
            col += 1
            value = sum([person_info['total_credit'] for person_info in group_data])
            school_credit.append(value)
            school_credit_girls.append(sum([person_info['total_credit'] for person_info in group_data if person_info['gender'] == 'female']))
            school_credit_boys.append(sum([person_info['total_credit'] for person_info in group_data if person_info['gender'] == 'male']))
            ws.write(row, col, value or '', self.easyxf(
                'font: bold on, color black; '
                'pattern: pattern solid, fore_color white; '
                'borders: top no_line, right no_line, bottom no_line, left no_line; '
            ))
            col += 1
            value = sum([person_info['balance'] for person_info in group_data])
            school_balance.append(value)
            school_balance_girls.append(sum([person_info['balance'] for person_info in group_data if person_info['gender'] == 'female']))
            school_balance_boys.append(sum([person_info['balance'] for person_info in group_data if person_info['gender'] == 'male']))
            ws.write(row, col, value or '', self.easyxf(
                'font: bold on, color black; '
                'pattern: pattern solid, fore_color white; '
                'borders: top no_line, right no_line, bottom no_line, left no_line; '
            ))
            col += 1
            if school_debit[-1]:
                value = self.div(float(school_credit[-1]),
                                 float(school_debit[-1]))
                ws.write(row, col, value, self.easyxf(
                    'font: bold on, color black; '
                    'pattern: pattern solid, fore_color %s; '
                    'borders: top no_line, right no_line, bottom no_line, left no_line; ' % self.get_background_color(value * 100), num_format_str='0%'
                ))
            else:
                ws.write(row, col, '', self.easyxf(
                    'font: bold on, color black; '
                    'pattern: pattern solid, fore_color white; '
                    'borders: top no_line, right no_line, bottom no_line, left no_line; '
                ))
            col += 1
            value = sum([person_info['previous'] for person_info in group_data])
            school_previous.append(value)
            school_previous_girls.append(sum([person_info['previous'] for person_info in group_data if person_info['gender'] == 'female']))
            school_previous_boys.append(sum([person_info['previous'] for person_info in group_data if person_info['gender'] == 'male']))
            ws.write(row, col, value or '', self.easyxf(
                'font: bold off, color black; '
                'pattern: pattern solid, fore_color white; '
                'borders: top no_line, right no_line, bottom no_line, left no_line; '
            ))
            row += 1

        title = _('Whole school')
        ws.write(row, 0, title, self.easyxf(
            'font: bold on, color black; '
            'pattern: pattern solid, fore_color white; '
            'borders: top medium, right no_line, bottom medium, left no_line; '
        ))
        col = 1
        for fee in self.collected_fees:
            ws.write(row, col, '', self.easyxf(
                'font: bold off, color black; '
                'pattern: pattern solid, fore_color white; '
                'borders: top medium, right no_line, bottom medium, left no_line; '))
            col += 1
        debit = sum(school_debit)
        ws.write(row, col, debit or '', self.easyxf(
            'font: bold on, color black; '
            'pattern: pattern solid, fore_color white; '
            'borders: top medium, right no_line, bottom medium, left no_line; '
        ))
        col += 1
        credit = sum(school_credit)
        ws.write(row, col, credit or '', self.easyxf(
            'font: bold on, color black; '
            'pattern: pattern solid, fore_color white; '
            'borders: top medium, right no_line, bottom medium, left no_line; '
        ))
        col += 1
        value = sum(school_balance)
        ws.write(row, col, value or '', self.easyxf(
            'font: bold on, color black; '
            'pattern: pattern solid, fore_color white; '
            'borders: top medium, right no_line, bottom medium, left no_line; '
        ))
        col += 1
        if debit:
            value = self.div(float(credit), float(debit))
            ws.write(row, col, value, self.easyxf(
                'font: bold on, color black; '
                'pattern: pattern solid, fore_color %s; '
                'borders: top medium, right no_line, bottom medium, left no_line; ' % self.get_background_color(value * 100), num_format_str='0%'
            ))
        else:
            ws.write(row, col, '', self.easyxf(
                'font: bold on, color black; '
                'pattern: pattern solid, fore_color white; '
                'borders: top no_line, right no_line, bottom no_line, left no_line; '
            ))
        col += 1
        value = sum(school_previous)
        ws.write(row, col, value or '', self.easyxf(
            'font: bold on, color black; '
            'pattern: pattern solid, fore_color white; '
            'borders: top medium, right medium, bottom medium, left no_line; '
        ))

        row += 1
        title = _('Girls')
        ws.write(row, 0, title, self.easyxf(
            'font: bold off, color black; '
            'pattern: pattern solid, fore_color white; '
            'borders: top no_line, right no_line, bottom no_line, left no_line; '
        ))
        col = 1
        for fee in self.collected_fees:
            ws.write(row, col, '', self.easyxf(
                'font: bold off, color black; '
                'pattern: pattern solid, fore_color white; '
                'borders: top no_line, right no_line, bottom no_line, left no_line; '))
            col += 1
        debit = sum(school_debit_girls)
        ws.write(row, col, debit or '', self.easyxf(
            'font: bold off, color black; '
            'pattern: pattern solid, fore_color white; '
            'borders: top no_line, right no_line, bottom no_line, left no_line; '
        ))
        col += 1
        credit = sum(school_credit_girls)
        ws.write(row, col, credit or '', self.easyxf(
            'font: bold off, color black; '
            'pattern: pattern solid, fore_color white; '
            'borders: top no_line, right no_line, bottom no_line, left no_line; '
        ))
        col += 1
        value = sum(school_balance_girls)
        ws.write(row, col, value or '', self.easyxf(
            'font: bold off, color black; '
            'pattern: pattern solid, fore_color white; '
            'borders: top no_line, right no_line, bottom no_line, left no_line; '
        ))
        col += 1
        if debit:
            value = self.div(float(credit), float(debit))
            ws.write(row, col, value, self.easyxf(
                'font: bold off, color black; '
                'pattern: pattern solid, fore_color %s; '
                'borders: top no_line, right no_line, bottom no_line, left no_line; ' % self.get_background_color(value * 100), num_format_str='0%'
            ))
        else:
            ws.write(row, col, '', self.easyxf(
                'font: bold on, color black; '
                'pattern: pattern solid, fore_color white; '
                'borders: top no_line, right no_line, bottom no_line, left no_line; '
            ))
        col += 1
        value = sum(school_previous_girls)
        ws.write(row, col, value or '', self.easyxf(
            'font: bold off, color black; '
            'pattern: pattern solid, fore_color white; '
            'borders: top no_line, right no_line, bottom no_line, left no_line; '
        ))
        row += 1
        title = _('Boys')
        ws.write(row, 0, title, self.easyxf(
            'font: bold off, color black; '
            'pattern: pattern solid, fore_color white; '
            'borders: top no_line, right no_line, bottom no_line, left no_line; '
        ))
        col = 1
        for fee in self.collected_fees:
            ws.write(row, col, '', self.easyxf(
                'font: bold off, color black; '
                'pattern: pattern solid, fore_color white; '
                'borders: top no_line, right no_line, bottom no_line, left no_line; '))
            col += 1
        debit = sum(school_debit_boys)
        ws.write(row, col, debit or '', self.easyxf(
            'font: bold off, color black; '
            'pattern: pattern solid, fore_color white; '
            'borders: top no_line, right no_line, bottom no_line, left no_line; '
        ))
        col += 1
        credit = sum(school_credit_boys)
        ws.write(row, col, credit or '', self.easyxf(
            'font: bold off, color black; '
            'pattern: pattern solid, fore_color white; '
            'borders: top no_line, right no_line, bottom no_line, left no_line; '
        ))
        col += 1
        value = sum(school_balance_boys)
        ws.write(row, col, value or '', self.easyxf(
            'font: bold off, color black; '
            'pattern: pattern solid, fore_color white; '
            'borders: top no_line, right no_line, bottom no_line, left no_line; '
        ))
        col += 1
        if debit:
            value = self.div(float(credit), float(debit))
            ws.write(row, col, value, self.easyxf(
                'font: bold off, color black; '
                'pattern: pattern solid, fore_color %s; '
                'borders: top no_line, right no_line, bottom no_line, left no_line; ' % self.get_background_color(value * 100), num_format_str='0%'
            ))
        else:
            ws.write(row, col, '', self.easyxf(
                'font: bold on, color black; '
                'pattern: pattern solid, fore_color white; '
                'borders: top no_line, right no_line, bottom no_line, left no_line; '
            ))
        col += 1
        value = sum(school_previous_boys)
        ws.write(row, col, value or '', self.easyxf(
            'font: bold off, color black; '
            'pattern: pattern solid, fore_color white; '
            'borders: top no_line, right no_line, bottom no_line, left no_line; '
        ))
        row += 2
        ws.write(row, 0, _('*YTD = Year to date'))

        row += 2
        title = _('Overview for selected stream')
        ws.write_merge(row, row, 0, self.merge, title, self.easyxf(
            'font: bold on, color white; '
            'pattern: pattern solid, fore_color report_title_bg; '
            'borders: top no_line, right no_line, bottom no_line, left no_line; '
        ))
        row += 1
        title = _('Fees owed')
        ws.write_merge(row, row, 1, self.merge - 4, title, self.easyxf(
            'font: bold on, color white; '
            'pattern: pattern solid, fore_color owed_bg; '
            'alignment: horizontal center; '
            'borders: top no_line, right no_line, bottom thin, left no_line; '
        ))
        ws.write(row, self.merge - 3, _('Fees paid'), self.easyxf(
            'font: bold on, color white; '
            'pattern: pattern solid, fore_color paid_bg; '
            'alignment: horizontal center; '
            'borders: top no_line, right no_line, bottom thin, left no_line; '
        ))
        title = _('Outstanding balance')
        ws.write_merge(row, row, self.merge - 2, self.merge, title, self.easyxf(
            'font: bold on, color white; '
            'pattern: pattern solid, fore_color balance_bg; '
            'alignment: horizontal center; '
            'borders: top no_line, right no_line, bottom thin, left no_line; '
        ))
        row += 1
        ws.write(row, 0, _('Name'), self.easyxf(
            'font: bold on, color black; '
            'pattern: pattern solid, fore_color white; '
            'borders: top no_line, right no_line, bottom thin, left no_line; '
        ))
        col = 1
        for fee in self.collected_fees:
            ws.write(row, col, fee.label, self.easyxf(
                'font: bold on, color white; '
                'pattern: pattern solid, fore_color owed_bg; '
                'alignment: horizontal center; '
                'borders: top no_line, right no_line, bottom thin, left no_line; '
            ))
            col += 1
        ws.write(row, col, _('Total YTD*'), self.easyxf(
            'font: bold on, color white; '
            'pattern: pattern solid, fore_color owed_bg; '
            'alignment: horizontal center; '
            'borders: top no_line, right no_line, bottom thin, left no_line; '
        ))
        col += 1
        ws.write(row, col, _('Total YTD*'), self.easyxf(
            'font: bold on, color white; '
            'pattern: pattern solid, fore_color paid_bg; '
            'alignment: horizontal center; '
            'borders: top no_line, right no_line, bottom thin, left no_line; '
        ))
        col += 1
        ws.write(row, col, _('Bal - YTD*'), self.easyxf(
            'font: bold on, color white; '
            'pattern: pattern solid, fore_color balance_bg; '
            'alignment: horizontal center; '
            'borders: top no_line, right no_line, bottom thin, left no_line; '
        ))
        col += 1
        ws.write(row, col, _('% paid'), self.easyxf(
            'font: bold on, color white; '
            'pattern: pattern solid, fore_color balance_bg; '
            'alignment: horizontal center; '
            'borders: top no_line, right no_line, bottom thin, left no_line; '
        ))
        col += 1
        ws.write(row, col, _('Prev. years'), self.easyxf(
            'font: bold on, color white; '
            'pattern: pattern solid, fore_color balance_bg; '
            'alignment: horizontal center; '
            'borders: top no_line, right no_line, bottom thin, left no_line; '
        ))
        row += 1
        total_group = {}
        total_girls = {}
        total_boys = {}
        for person_info in self.data:
            col = 0
            ws.write(row, col, person_info['title'], self.easyxf(
                'font: bold off, color black; '
                'pattern: pattern solid, fore_color white; '
                'borders: top no_line, right no_line, bottom no_line, left no_line; '
        ))
            col += 1
            for i, fee in enumerate(self.collected_fees):
                person_fee = person_info['fees'].get(fee.__name__)
                if person_fee is not None:
                    amount = person_fee.amount
                    ws.write(row, col, amount, self.easyxf(
                        'font: bold off, color black; '
                        'pattern: pattern solid, fore_color white; '
                        'borders: top no_line, right no_line, bottom no_line, left no_line; '
                    ))
                    if fee.__name__ not in total_group:
                        total_group[fee.__name__] = []
                    total_group[fee.__name__].append(amount)
                    if person_info['gender'] == 'male':
                        if fee.__name__ not in total_boys:
                            total_boys[fee.__name__] = []
                        total_boys[fee.__name__].append(amount)
                    else:
                        if fee.__name__ not in total_girls:
                            total_girls[fee.__name__] = []
                        total_girls[fee.__name__].append(amount)
                else:
                    ws.write(row, col, '', self.easyxf(
                        'font: bold off, color black; '
                        'pattern: pattern solid, fore_color white; '
                        'borders: top no_line, right no_line, bottom no_line, left no_line; '
                    ))
                col += 1
            ws.write(row, col, person_info['total_debit'] or '', self.easyxf(
                'font: bold on, color black; '
                'pattern: pattern solid, fore_color white; '
                'borders: top no_line, right no_line, bottom no_line, left no_line; '
            ))
            col += 1
            ws.write(row, col, person_info['total_credit'] or '', self.easyxf(
                'font: bold on, color black; '
                'pattern: pattern solid, fore_color white; '
                'borders: top no_line, right no_line, bottom no_line, left no_line; '
            ))
            col += 1
            ws.write(row, col, person_info['balance'] or '', self.easyxf(
                'font: bold on, color black; '
                'pattern: pattern solid, fore_color white; '
                'borders: top no_line, right no_line, bottom no_line, left no_line; '
            ))
            col += 1
            if person_info['total_debit']:
                value = self.div(float(person_info['total_credit']),
                                 float(person_info['total_debit']))
                ws.write(
                    row, col,
                    value,
                    self.easyxf(
                        'font: bold on, color black; '
                        'pattern: pattern solid, fore_color %s; '
                        'borders: top no_line, right no_line, bottom no_line, left no_line; ' % self.get_background_color(value * 100), num_format_str='0%'
                    ))
            else:
                ws.write(row, col, '', self.easyxf(
                    'font: bold on, color black; '
                    'pattern: pattern solid, fore_color white; '
                    'borders: top no_line, right no_line, bottom no_line, left no_line; '
                ))
            col += 1
            ws.write(row, col, person_info['previous'] or '', self.easyxf(
                'font: bold on, color black; '
                'pattern: pattern solid, fore_color white; '
                'borders: top no_line, right no_line, bottom no_line, left no_line; '
            ))
            row += 1
        title = _('Total for selected class')
        ws.write(row, 0, title, self.easyxf(
            'font: bold on, color black; '
            'pattern: pattern solid, fore_color white; '
            'borders: top medium, right no_line, bottom medium, left no_line; '
        ))
        col = 1
        for fee in self.collected_fees:
            ws.write(
                row, col, sum(total_group.get(fee.__name__, [])) or '', self.easyxf(
                    'font: bold off, color black; '
                    'pattern: pattern solid, fore_color white; '
                    'borders: top medium, right no_line, bottom medium, left no_line; '))
            col += 1
        debit = sum([person_info['total_debit'] for person_info in self.data])
        ws.write(row, col, debit or '', self.easyxf(
            'font: bold on, color black; '
            'pattern: pattern solid, fore_color white; '
            'borders: top medium, right no_line, bottom medium, left no_line; '
        ))
        col += 1
        credit = sum([person_info['total_credit'] for person_info in self.data])
        ws.write(row, col, credit or '', self.easyxf(
            'font: bold on, color black; '
            'pattern: pattern solid, fore_color white; '
            'borders: top medium, right no_line, bottom medium, left no_line; '
        ))
        col += 1
        value = sum([person_info['balance'] for person_info in self.data])
        ws.write(row, col, value or '', self.easyxf(
            'font: bold on, color black; '
            'pattern: pattern solid, fore_color white; '
            'borders: top medium, right no_line, bottom medium, left no_line; '
        ))
        col += 1
        if debit:
            value = self.div(float(credit), float(debit))
            ws.write(row, col, value, self.easyxf(
                'font: bold on, color black; '
                'pattern: pattern solid, fore_color %s; '
                'borders: top medium, right no_line, bottom medium, left no_line; ' % self.get_background_color(value * 100), num_format_str='0%'
            ))
        else:
            ws.write(row, col, '', self.easyxf(
                'font: bold on, color black; '
                'pattern: pattern solid, fore_color white; '
                'borders: top no_line, right no_line, bottom no_line, left no_line; '
            ))
        col += 1
        value = sum([person_info['previous'] for person_info in self.data])
        ws.write(row, col, value or '', self.easyxf(
            'font: bold on, color black; '
            'pattern: pattern solid, fore_color white; '
            'borders: top medium, right medium, bottom medium, left no_line; '
        ))

        row += 1
        title = _('Girls')
        ws.write(row, 0, title, self.easyxf(
            'font: bold off, color black; '
            'pattern: pattern solid, fore_color white; '
            'borders: top no_line, right no_line, bottom no_line, left no_line; '
        ))
        col = 1
        for fee in self.collected_fees:
            ws.write(
                row, col, sum(total_girls.get(fee.__name__, [])) or '', self.easyxf(
                    'font: bold off, color black; '
                    'pattern: pattern solid, fore_color white; '
                    'borders: top no_line, right no_line, bottom no_line, left no_line; '))
            col += 1
        debit = sum([person_info['total_debit'] for person_info in self.data if person_info['gender'] == 'female'])
        ws.write(row, col, debit or '', self.easyxf(
            'font: bold off, color black; '
            'pattern: pattern solid, fore_color white; '
            'borders: top no_line, right no_line, bottom no_line, left no_line; '
        ))
        col += 1
        credit = sum([person_info['total_credit'] for person_info in self.data if person_info['gender'] == 'female'])
        ws.write(row, col, credit or '', self.easyxf(
            'font: bold off, color black; '
            'pattern: pattern solid, fore_color white; '
            'borders: top no_line, right no_line, bottom no_line, left no_line; '
        ))
        col += 1
        value = sum([person_info['balance'] for person_info in self.data if person_info['gender'] == 'female'])
        ws.write(row, col, value or '', self.easyxf(
            'font: bold off, color black; '
            'pattern: pattern solid, fore_color white; '
            'borders: top no_line, right no_line, bottom no_line, left no_line; '
        ))
        col += 1
        if debit:
            value = self.div(float(credit), float(debit))
            ws.write(row, col, value, self.easyxf(
                'font: bold off, color black; '
                'pattern: pattern solid, fore_color %s; '
                'borders: top no_line, right no_line, bottom no_line, left no_line; ' % self.get_background_color(value * 100), num_format_str='0%'
            ))
        else:
            ws.write(row, col, '', self.easyxf(
                'font: bold on, color black; '
                'pattern: pattern solid, fore_color white; '
                'borders: top no_line, right no_line, bottom no_line, left no_line; '
            ))
        col += 1
        value = sum([person_info['previous'] for person_info in self.data if person_info['gender'] == 'female'])
        ws.write(row, col, value or '', self.easyxf(
            'font: bold off, color black; '
            'pattern: pattern solid, fore_color white; '
            'borders: top no_line, right no_line, bottom no_line, left no_line; '
        ))
        row += 1
        title = _('Boys')
        ws.write(row, 0, title, self.easyxf(
            'font: bold off, color black; '
            'pattern: pattern solid, fore_color white; '
            'borders: top no_line, right no_line, bottom no_line, left no_line; '
        ))
        col = 1
        for fee in self.collected_fees:
            ws.write(
                row, col, sum(total_boys.get(fee.__name__, [])) or '', self.easyxf(
                    'font: bold off, color black; '
                    'pattern: pattern solid, fore_color white; '
                    'borders: top no_line, right no_line, bottom no_line, left no_line; '))
            col += 1
        debit = sum([person_info['total_debit'] for person_info in self.data if person_info['gender'] == 'male'])
        ws.write(row, col, debit or '', self.easyxf(
            'font: bold off, color black; '
            'pattern: pattern solid, fore_color white; '
            'borders: top no_line, right no_line, bottom no_line, left no_line; '
        ))
        col += 1
        credit = sum([person_info['total_credit'] for person_info in self.data if person_info['gender'] == 'male'])
        ws.write(row, col, credit or '', self.easyxf(
            'font: bold off, color black; '
            'pattern: pattern solid, fore_color white; '
            'borders: top no_line, right no_line, bottom no_line, left no_line; '
        ))
        col += 1
        value = sum([person_info['balance'] for person_info in self.data if person_info['gender'] == 'male'])
        ws.write(row, col, value or '', self.easyxf(
            'font: bold off, color black; '
            'pattern: pattern solid, fore_color white; '
            'borders: top no_line, right no_line, bottom no_line, left no_line; '
        ))
        col += 1
        if debit:
            value = self.div(float(credit), float(debit))
            ws.write(row, col, value, self.easyxf(
                'font: bold off, color black; '
                'pattern: pattern solid, fore_color %s; '
                'borders: top no_line, right no_line, bottom no_line, left no_line; ' % self.get_background_color(value * 100), num_format_str='0%'
            ))
        else:
            ws.write(row, col, '', self.easyxf(
                'font: bold on, color black; '
                'pattern: pattern solid, fore_color white; '
                'borders: top no_line, right no_line, bottom no_line, left no_line; '
            ))
        col += 1
        value = sum([person_info['previous'] for person_info in self.data if person_info['gender'] == 'male'])
        ws.write(row, col, value or '', self.easyxf(
            'font: bold off, color black; '
            'pattern: pattern solid, fore_color white; '
            'borders: top no_line, right no_line, bottom no_line, left no_line; '
        ))
        row += 2
        ws.write(row, 0, _('*YTD = Year to date'))
        self.finish('sheet1')

    def __call__(self):
        self.makeProgress()
        self.task_progress.title = _("Generating report")
        self.addExporters(self.task_progress)

        wb = xlwt.Workbook()
        self.setup_book(wb)
        self.task_progress.force('data', active=True)
        self.data
        self.finish('data')

        self.export_report(wb)

        self.task_progress.title = _("Report complete")
        self.task_progress.force('overall', progress=1.0)
        if 'DEBUG' in self.request:
            return self.renderize(wb)
        else:
            return wb

    def renderize(self, workbook):
        datafile = StringIO()
        workbook.save(datafile)
        data = datafile.getvalue()
        self.setUpHeaders(data)
        return data

    def print_legend_1(self, ws, row=None, col=None):
        labels = [
            (_('85-100%'), 90),
            (_('70-85%'), 80),
            (_('50-70%'), 60),
            (_('Under 50%'), 0),
        ]
        style = self.easyxf(
            'font: bold on, color black; '
            'pattern: pattern solid, fore_color white; '
            'alignment: horizontal center, vertical center, wrap on; '
            'borders: top medium, right medium, bottom medium, left medium; '
        )
        ws.write_merge(row, row + 3, col, col, _('Key for % fees paid'), style)
        borders = self.translate_borders([1, 1, 1, 1])
        col += 1
        for i, label_info in enumerate(labels):
            label, value = label_info
            background = self.get_background_color(value)
            style = self.easyxf(
                'font: bold on, color black; '
                'pattern: pattern solid, fore_color %s; '
                'alignment: horizontal center; '
                'borders: %s' % (background, borders),
                num_format_str='0')
            ws.write(row, col, label, style)
            row += 1

    @Lazy
    def custom_colors(self):
        result = {
            'above85': (60, 146, 208, 80),
            'above70': (61, 255, 255, 102),
            'above50': (62, 247, 150, 70),
            'below50': (63, 255, 0, 0),
            'owed_bg': (57, 49, 133, 156),
            'paid_bg': (58, 149, 55, 53),
            'balance_bg': (59, 128, 128, 128),
            'report_title_bg': (55, 31, 73, 125),
            'table_title_bg': (56, 217, 217, 217),
        }
        return result

    def setup_book(self, wb):
        for color, info in self.custom_colors.items():
            code, r, g, b = info
            xlwt.add_palette_colour(color, code)
            wb.set_colour_RGB(code, r, g, b)

    def get_background_color(self, value):
        if value >= 85:
            return 'above85'
        elif value >= 70:
            return 'above70'
        elif value >= 50:
            return 'above50'
        else:
            return 'below50'

    @Lazy
    def styles(self):
        result = {}
        pattern = 'pattern: pattern solid, fore_colour %s'
        for color in self.custom_colors:
            result[color] = self.easyxf(pattern % color)
        result.update({
            'block_top_row': self.easyxf(
                'font: bold off, color black; '
                'pattern: pattern solid, fore_color white; '
                'borders: top thin, right thin, bottom no_line, left no_line; '
            ),
            'block_bottom_row': self.easyxf(
                'font: bold off, color black; '
                'pattern: pattern solid, fore_color white; '
                'borders: top no_line, right thin, bottom thin, left no_line; '
            ),
            'table_title': self.easyxf(
                'font: bold on, color black; '
                'pattern: pattern solid, fore_color table_title_bg; '
                'borders: top no_line, right no_line, bottom no_line, left no_line; '
            ),
            'column_header': self.easyxf(
                'font: bold on, color black; '
                'pattern: pattern solid, fore_color white; '
                'borders: top thin, right thin, bottom thin, left thin; '
                'alignment: horizontal center; '
            ),
        })
        return result

    @Lazy
    def group(self):
        groups = IStreamContainer(self.schoolyear)
        return groups.get(self.request.get('group'))

    @Lazy
    def schoolyear(self):
        return self.context

    @Lazy
    def schoolyear_fees(self):
        return IFeeContainer(self.schoolyear)

    def person_total_debit(self, fees, first, last):
        return sum([fee.amount for fee in fees.values()
                    if first <= fee.date <= last])

    def person_total_credit(self, person, first, last):
        fees = IFeeContainer(person)
        return sum([fee.amount for fee in fees.values()
                    if ICredit.providedBy(fee) and
                    first <= fee.date <= last])

    @Lazy
    def data(self):
        result = []
        today = getUtility(IDateManager).today
        for person in sorted(self.group.members,
                             key=lambda p: (p.last_name, p.first_name)):
            fees = {}
            person_fees = IFeeContainer(person)
            for fee in self.schoolyear_fees.values():
                if fee.__name__ in person_fees and fee.date <= today:
                    fees[fee.__name__] = person_fees[fee.__name__]
            total_debit = self.person_total_debit(
                fees, self.schoolyear.first, today)
            total_credit = self.person_total_credit(
                person, self.schoolyear.first, today)
            info = {
                'gender': person.gender,
                'title': person.title,
                'fees': fees,
                'total_debit': total_debit,
                'total_credit': total_credit,
                'balance': total_credit - total_debit,
                'previous': 0,
            }
            result.append(info)
        return result

    def div(self, count, total):
        if not total:
            return 0.0
        return count/float(total)

    @Lazy
    def merge(self):
        return len(self.collected_fees) + 5

    @Lazy
    def collected_fees(self):
        unique = set()
        for person_info in self.data:
            for person_fee in person_info['fees'].values():
                fee = self.schoolyear_fees.get(person_fee.__name__)
                if fee is not None:
                    unique.add(fee)
        return sorted(unique, key=lambda f: f.title)

    def print_headers(self, ws, title):
        row = 0
        self.print_report_title(ws, title, row=row, merge=self.merge)

        row = 2
        today = getUtility(IDateManager).today
        self.print_report_date(ws, today, row=row)

        row += 1
        self.print_report_filters(ws, row=row)

    def print_report_title(self, ws, title, row, merge):
        style = self.easyxf(
            'font: bold on, color white; '
            'pattern: pattern solid, fore_color report_title_bg; '
        )
        ws.write_merge(row, row, 0, merge, title, style)

    def print_report_date(self, ws, today, row):
        style = self.easyxf(
            'font: bold on, color black; '
            'pattern: pattern solid, fore_color white; '
            'borders: top medium, right no_line, bottom no_line, left medium; '
        )
        ws.write(row, 0, _('Date of report'), style)
        style = self.easyxf(
            'font: bold on, color black; '
            'pattern: pattern solid, fore_color white; '
            'borders: top medium, right medium, bottom no_line, left no_line; '
        )
        ws.write(row, 1, today.strftime('%d/%m/%Y'), style)

    def translate_borders(self, borders):
        result = []
        order = ['top', 'right', 'bottom', 'left']
        for i, enabled in  enumerate(borders):
            result.append('%s %s' % (order[i], ['no_line', 'thin'][enabled]))
        return ', '.join(result)        

    def print_report_filters(self, ws, row):
        style = self.easyxf(
            'font: bold on, color black; '
            'pattern: pattern solid, fore_color white; '
            'borders: top no_line, right no_line, bottom medium, left medium; '
        )
        ws.write(row, 0, STMessage('Stream'), style)
        style = self.easyxf(
            'font: bold on, color black; '
            'pattern: pattern solid, fore_color white; '
            'borders: top no_line, right medium, bottom medium, left no_line; '
        )
        ws.write(row, 1, self.group.title, style)


class CreditImportFormReportRequestView(RequestXLSReportDialog):

    report_builder = 'credit_import_form.xls'


class CreditImportFormReportView(ExcelExportView):

    message_title = _('Fees report: Credit import form')

    sheet_1_label = _('Credit import form')

    overall_line_id = 'overall'
    _style_cache = {}

    def easyxf(self, s, num_format_str=None):
        key = s.strip(), num_format_str
        if key not in self._style_cache:
            self._style_cache[key] = xlwt.easyxf(s, num_format_str)
        return self._style_cache[key]

    def addExporters(self, progress):
        progress.add(
            'overall', title=_('Report'), progress=0.0)
        progress.add(
            'data', title=_('Gathering report data'),
            active=False, progress=0.0)

    @property
    def base_filename(self):
        return 'credit_import_form_%s' % self.schoolyear.title

    def print_headers(self, ws):
        row = 0
        headers = [
            'User ID',
            'Class',
            'First name',
            'Last name',
            'Gender',
            'Credit date',
            'Title',
            'Description',
            'Amount',
            'Receipt number',
        ]
        for i, header in enumerate(headers):
            self.write_header(ws, row, i, header)

    def export_report(self, wb):
        self.task_progress.force('sheet1', active=True)
        ws = wb.add_sheet(self.sheet_1_label)
        self.print_headers(ws)
        row = 1
        for info in self.data:
            ws.write(row, 0, info['username'])
            ws.write(row, 1, info['level'])
            ws.write(row, 2, info['first_name'])
            ws.write(row, 3, info['last_name'])
            ws.write(row, 4, info['gender'])
            row += 1
        self.finish('sheet1')

    def setup_book(self, wb):
        pass

    def __call__(self):
        self.makeProgress()
        self.task_progress.title = _("Generating report")
        self.addExporters(self.task_progress)

        wb = xlwt.Workbook()
        self.setup_book(wb)
        self.task_progress.force('data', active=True)
        self.data
        self.finish('data')

        self.export_report(wb)

        self.task_progress.title = _("Report complete")
        self.task_progress.force('overall', progress=1.0)
        if 'DEBUG' in self.request:
            return self.renderize(wb)
        else:
            return wb

    def renderize(self, workbook):
        datafile = StringIO()
        workbook.save(datafile)
        data = datafile.getvalue()
        self.setUpHeaders(data)
        return data

    @Lazy
    def group(self):
        groups = IGroupContainer(self.schoolyear)
        return groups['students']

    @Lazy
    def schoolyear(self):
        return self.context

    @Lazy
    def data(self):
        result = []
        for person in sorted(self.group.members,
                             key=lambda p: (p.last_name, p.first_name)):
            info = {
                'username': person.username,
                'level': self.get_level(person),
                'first_name': person.first_name,
                'last_name': person.last_name,
                'gender': person.gender,
            }
            result.append(info)
        return sorted(result, key=lambda i: (i['level'], i['first_name']))

    def get_level(self, person):
        result = []
        levels = person.levels.all()
        for level in levels.any(ACTIVE):
            result.append(level)
        if result:
            return result[0].title
        return ''


class ImportCreditLink(flourish.page.LinkViewlet,
                       ActiveSchoolYearContentMixin):

    @property
    def url(self):
        url = absoluteURL(self.schoolyear, self.request)
        return '%s/import_credits.html' % url


class CreditImporter(ImporterBase):

    title = _('Credits')
    sheet_name = 'Credit import form'

    def create_credit(self, data):
        result = Credit()
        result.date = data['date']
        result.title = data['title']
        result.amount = data['amount']
        result.description = data['description']
        result.receipt = data['receipt']
        return result

    def process(self):
        sh = self.sheet
        first_row = 1
        nrows = sh.nrows
        app = ISchoolToolApplication(None)
        numbers = IReceiptNumbers(app)
        persons = app['persons']
        for row in range(first_row, nrows):
            if self.isEmptyRow(sh, row):
                continue
            num_errors = len(self.errors)
            data = {}
            data['__name__'] = self.getRequiredIdFromCell(sh, row, 0)
            data['date'] = self.getDateFromCell(sh, row, 5)
            data['title'] = self.getTextFromCell(sh, row, 6)
            data['description'] = self.getTextFromCell(sh, row, 7)
            data['amount'] = self.getRequiredTextFromCell(sh, row, 8)
            data['receipt'] = self.getTextFromCell(sh, row, 9)
            if not data['receipt']:
                data['receipt'] = unicode(numbers)
            if data['__name__'] not in persons:
                self.error(row, 0, ERROR_INVALID_PERSON_ID)
            try:
                if data['amount']:
                    data['amount'] = Decimal(data['amount'])
                    if data['amount'] < 0:
                        raise InvalidOperation()
            except InvalidOperation:
                self.error(row, 8, _('has an invalid amount'))
            if num_errors == len(self.errors):
                person = persons[data['__name__']]
                fees = IFeeContainer(person)
                chooser = INameChooser(fees)
                credit = self.create_credit(data)
                name = chooser.chooseName(u'', credit)
                fees[name] = credit
            self.progress(row, nrows)


class ImportCreditImporter(RemoteMegaImporter):

    @property
    def importers(self):
        return [
            CreditImporter,
        ]


class ImportCreditView(FlourishRemoteMegaImporter):

    def getTask(self):
        app = ISchoolToolApplication(None)
        return ImportTask(ImportCreditImporter, app)


class SchoolYearFeeView(AmountView):

    template = InheritTemplate(flourish.page.Page.template)
    content_template = ViewPageTemplateFile('templates/schoolyear_fee.pt')

    def recurrences(self):
        result = []
        container = self.context.__parent__
        for item in container.values():
            if item.master == self.context.__name__:
                result.append(item)
        return sorted(result, key=lambda f: f.date)

    @Lazy
    def person(self):
        pass

    @property
    def fields(self):
        return field.Fields(IDebit).select(
            'date',
            'last_date',
            'label',
            'description',
            'amount')

    @property
    def canModify(self):
        return checkPermission('schooltool.edit', self.context)

    @property
    def title(self):
        return self.context.title

    def nextURL(self):
        return absoluteURL(self.context.__parent__, self.request)

    def edit_url(self):
        item_url = absoluteURL(self.context, self.request)
        return '%(url)s/edit_schoolyear_fee.html?camefrom=%(url)s' % {'url': item_url}


class GroupFeesViewlet(flourish.viewlet.Viewlet):

    title = _('Group Fees')

    template = flourish.templates.File('templates/group_fees_table.pt')

    @Lazy
    def fees(self):
        result = []
        for fee in IFeeContainer(self.context).values():
            if not fee.master:
                edit_url = '%s/edit_fee.html?fee=%s' % (
                    absoluteURL(self.context, self.request), fee.__name__)
                result.append({
                    'title': fee.title,
                    'edit_url': edit_url,
                })
        return sorted(result, key=lambda i: i['title'])


class GroupFeeContainerTable(table.ajax.Table):

    no_default_url_cell_formatter = True

    def columns(self):
        title = GetterColumn(
            name='title',
            title=STMessage(u'Title'),
            getter=lambda i, f: i.title)
        return [title]


class GroupEditFeesTable(BasicPersonTable):

    batch_size = 0

    def __init__(self, *args, **kw):
        super(GroupEditFeesTable, self).__init__(*args, **kw)
        self.errors = {}

    @property
    def search_fee_id(self):
        return self.html_id + '-fee'

    @property
    def action_name(self):
        return self.html_id + '-action'

    @property
    def submit_name(self):
        return self.html_id + '-submit'

    @property
    def cancel_name(self):
        return '___CANCEL___'

    @property
    def input_prefix(self):
        return self.html_id + '-input'

    def items(self):
        return self.indexItems(self.context.members)

    @Lazy
    def selected_fee(self):
        if 'fee' in self.request:
            requested_fee = self.request.get('fee')
        else:
            requested_fee = self.request.get(self.search_fee_id)
        return IFeeContainer(self.context).get(requested_fee)

    def nextURL(self):
        return absoluteURL(self.context, self.request)

    def columns(self):
        default = BasicPersonTable.columns(self)
        checkbox = FeeCheckboxColumn(
            self.action_name,
            name='checkbox',
            title=_('Apply?'),
            id_getter=lambda i: i.__name__)
        amount = FeeAmountInputColumn(
            self.input_prefix,
            name='amount',
            title=_('Amount'))
        weakness = GetterColumn(
            name='weakness',
            title=_('EWS / Freeship'),
            getter=lambda i, f: IDemographics(i).get('economic_weakness') or '')
        result = [checkbox, amount] + default
        if self.add_weakness_column:
            result = result + [weakness]
        return result

    @Lazy
    def add_weakness_column(self):
        app = ISchoolToolApplication(None)
        dfs = IDemographicsFields(app)
        return 'economic_weakness' in dfs

    def update(self):
        super(GroupEditFeesTable, self).update()
        view_url = '%s/edit_fee.html' % absoluteURL(self.context, self.request)
        if self.cancel_name in self.request:
            url = absoluteURL(self.context, self.request)
            self.request.response.redirect(url)
            return
        if self.submit_name in self.request:
            if self.selected_fee is not None:
                submitted_persons = list(self.request.get(self.action_name, []))
                errors = self.validate(submitted_persons)
                if not errors:
                    self.process(submitted_persons)
                    url = view_url + '?SUCCESS=1&fee=%s' % self.selected_fee.__name__
                    self.request.response.redirect(url)
                    return
                else:
                    self.request.response.redirect(
                        view_url + '?ERROR=1&fee=%s' % self.selected_fee.__name__)
                    return
            else:
                self.request.response.redirect(view_url)

    def validate(self, submitted_persons):
        for person_id in submitted_persons:
            value_name = input_name(self.input_prefix, person_id)
            value = convert_value(self.request.get(value_name))
            if value is None:
                self.errors[person_id] = None
        return self.errors

    def process(self, submitted_persons):
        for item in self._items:
            person = table.column.unindex(item)
            checked = person.__name__ in submitted_persons
            fee = IFeeContainer(person).get(self.selected_fee.__name__)
            if fee is not None:
                if checked:
                    value_name = input_name(self.input_prefix, person.__name__)
                    value = convert_value(self.request.get(value_name))
                    if value != fee.amount:
                        fee.amount = value
                        for instance in self.instances(self.selected_fee):
                            if instance.__name__ in IFeeContainer(person):
                                IFeeContainer(person)[instance.__name__].amount = value
                else:
                    del IFeeContainer(person)[self.selected_fee.__name__]
            else:
                if checked:
                    fee = Debit()
                    value_name = input_name(self.input_prefix, person.__name__)
                    fee.amount = convert_value(self.request.get(value_name))
                    fee.title = self.selected_fee.title
                    fee.date = self.selected_fee.date
                    fee.last_date = self.selected_fee.last_date
                    fee.label = self.selected_fee.label
                    fee.description = self.selected_fee.description
                    IFeeContainer(person)[self.selected_fee.__name__] = fee
                    for instance in self.instances(self.selected_fee):
                        if instance.__name__ in IFeeContainer(person):
                            recurrence = IFeeContainer(person)[instance.__name__]
                            if recurrence.amount != fee.amount:
                                recurrence.amount = fee.amount

    def recurrences(self, fee):
        result = []
        container = fee.__parent__
        for item in container.values():
            if item.master == fee.__name__:
                result.append(item)
        return result

    def instances(self, fee):
        result = []
        container = fee.__parent__
        if not fee.master:
            master = fee
        else:
            master = container[fee.master]
            result.append(master)
        for item in container.values():
            if (item.master == master.__name__ and
                item.__name__ != fee.__name__):
                result.append(item)
        return result


class GroupEditFeesTableFilter(table.ajax.TableFilter):

    template = flourish.templates.File('templates/group_edit_fee_table_filter.pt')

    def fees(self):
        result = []
        container = IFeeContainer(self.context)
        fee = self.manager.selected_fee
        if fee is not None:
            result.append({
                'id': fee.__name__,
                'title': fee.title,
                'date': fee.date,
            })
            if fee.master:
                master = container.get(fee.master)
                if master is not None:
                    result.append({
                        'id': master.__name__,
                        'title': master.title,
                        'date': master.date,
                    })
                for recurrence in self.recurrences(master):
                    if recurrence.__name__ == fee.__name__:
                        continue
                    result.append({
                        'id': recurrence.__name__,
                        'title': recurrence.title,
                        'date': recurrence.date,
                    })
            else:
                for recurrence in self.recurrences(fee):
                    result.append({
                        'id': recurrence.__name__,
                        'title': recurrence.title,
                        'date': recurrence.date,
                    })
        return sorted(result, key=lambda d: (d['date'], d['title']))

    def has_fee(self, item, fee):
        person = table.column.unindex(item)
        return fee.__name__ in IFeeContainer(person)

    def recurrences(self, fee):
        result = []
        container = fee.__parent__
        for item in container.values():
            if item.master == fee.__name__:
                result.append(item)
        return result


class GroupFeeEditView(flourish.page.Page):

    container_class = 'container widecontainer'

    def update(self):
        zc.resourcelibrary.need('schooltool.fee.flourish')

    @property
    def title(self):
        return self.context.title

    @Lazy
    def schoolyear_fees(self):
        schoolyear = ISchoolYear(self.context.__parent__)
        return IFeeContainer(schoolyear)


class SchoolYearFeeAddLinks(flourish.page.RefineLinksViewlet):

    pass


class IDailyFeeCollectionReport(Interface):

    date = Date(
        title=_('Day'))


class DailyFeeCollectionReportRequestView(RequestXLSReportDialog):

    report_builder = 'daily_fee_collection_report.xls'

    fields = field.Fields(IDailyFeeCollectionReport)

    def updateTaskParams(self, task):
        date = self.form_params.get('date')
        day = '%d-%02d-%02d' % (date.year, date.month, date.day)
        task.request_params['day'] = day


DailyReport_date = widget.ComputedWidgetAttribute(
    lambda adapter: adapter.request.util.today,
    view=DailyFeeCollectionReportRequestView,
    field=IDailyFeeCollectionReport['date']
    )


class DailyFeeCollectionReportView(BasicFeesReportView):

    message_title = _('Fees report: Daily fee collection report')
    sheet_1_title = _('Daily fee collection report')
    sheet_1_label = _('Daily collection report')

    @Lazy
    def day(self):
        day = self.request.get('day', None)
        year, month, day = [int(part) for part in day.split('-')]
        return datetime.date(year, month, day)

    @property
    def base_filename(self):
        return 'daily_fee_collection_%s' % self.day.strftime('%Y-%m-%d')

    def print_report_filters(self, ws, row):
        style = self.easyxf(
            'font: bold on, color black; '
            'pattern: pattern solid, fore_color white; '
            'borders: top no_line, right no_line, bottom medium, left medium; '
        )
        ws.write(row, 0, _('Date selected for data'), style)
        style = self.easyxf(
            'font: bold on, color black; '
            'pattern: pattern solid, fore_color white; '
            'borders: top no_line, right medium, bottom medium, left no_line; '
        )
        ws.write(row, 1, self.day.strftime('%d/%m/%Y'), style)

    def export_report(self, wb):
        self.task_progress.force('sheet1', active=True)
        ws = wb.add_sheet(self.sheet_1_label)
        self.print_headers(ws, self.sheet_1_title)
        row = 5
        headers = [
            'Serial number',
            'Class',
            'Name',
            'Student ID',
            'Fee paid on selected date',
            'Description',
            'Receipt number',
            'Fees outstanding',
        ]
        for i, header in enumerate(headers):
            ws.write(row, i, header, self.styles['column_header'])
        row += 1
        for i, info in enumerate(self.data):
            ws.write(row, 0, i+1)
            ws.write(row, 1, info['level'])
            ws.write(row, 2, info['person'].title)
            ws.write(row, 3, info['person'].username)
            ws.write(row, 4, info['fee'].amount)
            ws.write(row, 5, info['fee'].description)
            ws.write(row, 6, info['fee'].receipt)
            ws.write(row, 7, info['balance'])
            row += 1
        row += 1
        levels = sorted(set([info['level'] for info in self.data]))
        for level in levels:
            if not level:
                level = 'No Level'
            ws.write(row, 1, 'Total %s' % level)
            ws.write(row, 4, self.level_total(level))
            ws.write(row, 7, self.level_balance(level))
            row += 1
        style = self.easyxf(
            'font: bold on, color black; '
            'pattern: pattern solid, fore_color white; '
            'borders: top medium, right no_line, bottom medium, left no_line; '
        )
        ws.write(row, 1, 'Grand total for selected date', style)
        ws.write(row, 2, '', style)
        ws.write(row, 3, '', style)
        ws.write(row, 4, self.date_total(), style)
        ws.write(row, 5, '', style)
        ws.write(row, 6, '', style)
        ws.write(row, 7, self.date_balance(), style)
        self.finish('sheet1')

    def date_total(self):
        result = []
        for info in self.data:
            result.append(info['fee'].amount)
        return sum(result)

    def date_balance(self):
        result = []
        for info in self.data:
            result.append(info['balance'])
        return sum(result)

    def level_total(self, level):
        result = []
        for info in self.data:
            if info['level'] == level:
                result.append(info['fee'].amount)
        return sum(result)

    def level_balance(self, level):
        result = []
        for info in self.data:
            if info['level'] == level:
                result.append(info['balance'])
        return sum(result)

    def sortKey(self, info):
        level = info['level'] or ''
        return level, info['person'].title

    def level(self, person):
        result = []
        person = removeSecurityProxy(person)
        for level in person.levels.on(self.day).any(ACTIVE):
            result.append(level)
        if result:
            return result[0].title

    def balance(self, person):
        fees = {}
        today = getUtility(IDateManager).today
        person_fees = IFeeContainer(person)
        for fee in self.schoolyear_fees.values():
            if fee.__name__ in person_fees and fee.date <= today:
                fees[fee.__name__] = person_fees[fee.__name__]
        total_debit = self.person_total_debit(
            fees, self.schoolyear.first, today)
        total_credit = self.person_total_credit(
            person, self.schoolyear.first, today)
        return total_credit - total_debit

    @Lazy
    def merge(self):
        return 7

    @Lazy
    def data(self):
        result = []
        app = ISchoolToolApplication(None)
        persons = app['persons']
        for person in persons.values():
            fees = IFeeContainer(person)
            for fee in fees.values():
                if ICredit.providedBy(fee) and fee.date == self.day:
                    result.append({
                        'level': self.level(person),
                        'person': person,
                        'fee': fee,
                        'balance': self.balance(person),
                    })
        return sorted(result, key=lambda info: self.sortKey(info))


class AddGroupFeesToConvertedStreamSubscriber(ObjectEventAdapterSubscriber):

    adapts(IGroupConvertedToStreamEvent, IStream)

    def __call__(self):
        schoolyear = ISchoolYear(self.object)
        group = IGroupContainer(schoolyear).get(self.object.__name__)
        if group is not None:
            stream_fees = IFeeContainer(self.object)
            fees = IFeeContainer(group)
            for debit in fees.values():
                name = debit.__name__
                if name not in stream_fees:
                    stream_fee = Debit()
                    stream_fee.date = debit.date
                    stream_fee.last_date = debit.last_date
                    stream_fee.title = debit.title
                    stream_fee.label = debit.label
                    stream_fee.description = debit.description
                    stream_fee.amount = debit.amount
                    stream_fee.master = debit.master
                    stream_fees[name] = stream_fee
