#
# SchoolTool - common information systems platform for school administration
# Copyright (c) 2014 Shuttleworth Foundation
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"""
Income activities views
"""

from zope.browserpage.viewpagetemplatefile import ViewPageTemplateFile
from zope.component import adapts
from zope.component import getMultiAdapter
from zope.container.interfaces import INameChooser
from zope.i18n import translate
from zope.interface import implements
from zope.proxy import sameProxiedObjects
from zope.schema import Bool
from zope.security import checkPermission
from zope.security.proxy import removeSecurityProxy
from zope.traversing.browser.absoluteurl import absoluteURL
from zope.traversing.browser.interfaces import IAbsoluteURL

import zc.resourcelibrary
from z3c.form import button
from z3c.form import field
from z3c.form import form
from z3c.form import widget
from zc.table.column import GetterColumn

from schooltool import table
from schooltool.app.interfaces import ISchoolToolApplication
from schooltool.common import SchoolToolMessage as STMessage
from schooltool.common.inlinept import InheritTemplate
from schooltool.report.browser.report import RequestRemoteReportDialog
from schooltool.skin import flourish

from schooltool.fee.browser.fee import PersonFeeSummaryPDFView
from schooltool.fee.income import IncomeGeneratingActivity
from schooltool.fee.income import IncomeGeneratingActivityCategory
from schooltool.fee.interfaces import IIncomeGeneratingActivity
from schooltool.fee.interfaces import IIncomeGeneratingActivityCategory
from schooltool.fee.interfaces import IIncomeGeneratingActivityCategoryContainer
from schooltool.fee.interfaces import IIncomeGeneratingActivityContainer

from schooltool.fee import FeeMessage as _


class IncomeGeneratingActivityContainerAbsoluteURL(flourish.page.Page):

    adapts(IIncomeGeneratingActivityContainer,
           flourish.interfaces.IFlourishLayer)
    implements(IAbsoluteURL)

    def __str__(self):
        app = ISchoolToolApplication(None)
        url = str(getMultiAdapter((app, self.request), name='absolute_url'))
        return url + '/income_generating_activities'

    __call__ = __str__


class IncomeGeneratingActivityContainerView(flourish.page.Page):

    content_template = ViewPageTemplateFile(
        'templates/income_activity_container.pt')


class IncomeGeneratingActivityContainerTable(table.ajax.Table):

    pass


class IncomeGeneratingActivityContainerAddLinks(
        flourish.page.RefineLinksViewlet):

    pass


class IncomeGeneratingActivityContainerCustomizeLinks(
        flourish.page.RefineLinksViewlet):

    pass


field_order = ['title', 'description', 'category', 'projected_income']


class IncomeGeneratingActivityAddView(flourish.form.AddForm):

    template = InheritTemplate(flourish.page.Page.template)
    label = None
    legend = _('Income Generating Activity Information')
    fields = field.Fields(IIncomeGeneratingActivity).select(*field_order)

    def updateActions(self):
        super(IncomeGeneratingActivityAddView, self).updateActions()
        self.actions['add'].addClass('button-ok')
        self.actions['cancel'].addClass('button-cancel')

    @button.buttonAndHandler(STMessage('Submit'), name='add')
    def handleAdd(self, action):
        super(IncomeGeneratingActivityAddView, self).handleAdd.func(self, action)

    @button.buttonAndHandler(STMessage('Cancel'))
    def handleCancel(self, action):
        url = absoluteURL(self.context, self.request)
        self.request.response.redirect(url)

    def nextURL(self):
        return absoluteURL(self._activity, self.request)

    def create(self, data):
        activity = IncomeGeneratingActivity()
        form.applyChanges(self, activity, data)
        return activity

    def add(self, activity):
        chooser = INameChooser(self.context)
        name = chooser.chooseName(u'', activity)
        self.context[name] = activity
        self._activity = activity


def get_default_category(adapter):
    app = ISchoolToolApplication(None)
    return IIncomeGeneratingActivityCategoryContainer(app).default


AddIncomeGeneratingActivityDefaultCategory = widget.ComputedWidgetAttribute(
    get_default_category,
    view=IncomeGeneratingActivityAddView,
    field=IIncomeGeneratingActivity['category']
    )


class IncomeGeneratingActivityView(flourish.form.DisplayForm):

    template = InheritTemplate(flourish.page.Page.template)
    content_template = ViewPageTemplateFile('templates/income_activity.pt')
    fields = field.Fields(IIncomeGeneratingActivity).omit(
        'title', 'current_income')

    @property
    def canModify(self):
        return (not self.context.closed and
                checkPermission('schooltool.edit', self.context))

    @property
    def title(self):
        return self.context.title

    def nextURL(self):
        return absoluteURL(self.context.__parent__, self.request)

    def edit_url(self):
        return '%s/edit.html' % absoluteURL(self.context, self.request)

    def render(self, *args, **kw):
        zc.resourcelibrary.need('schooltool.fee.flourish')
        view = super(IncomeGeneratingActivityView, self)
        return view.render(*args, **kw)


class IncomeGeneratingActivityCategoryContainerView(flourish.page.Page):

    content_template = ViewPageTemplateFile(
        'templates/income_activity_category_container.pt')

    def render(self, *args, **kw):
        zc.resourcelibrary.need('schooltool.fee.flourish')
        view = super(IncomeGeneratingActivityCategoryContainerView, self)
        return view.render(*args, **kw)

    def nextURL(self):
        app = ISchoolToolApplication(None)
        container = IIncomeGeneratingActivityContainer(app)
        return absoluteURL(container, self.request)


class CheckColumn(GetterColumn):

    def getter(self, item, formatter):
        container = item.__parent__
        return sameProxiedObjects(item, container.default)

    def cell_formatter(self, value, item, formatter):
        if not value:
            return ''
        return '<span class="ui-icon ui-icon-check"></span>'


class IncomeGeneratingActivityCategoryContainerTable(table.ajax.Table):

    def columns(self):
        columns = super(IncomeGeneratingActivityCategoryContainerTable,
                        self).columns()
        default = CheckColumn(
            name='default',
            title=STMessage('Default'))
        return columns + [default]

    def updateFormatter(self):
        if self._table_formatter is None:
            self.setUp(table_formatter=self.table_formatter,
                       batch_size=self.batch_size,
                       prefix=self.__name__,
                       css_classes={'table': 'data income-activity-categories'})


class IncomeGeneratingActivityCategoryContainerAddLinks(
        flourish.page.RefineLinksViewlet):

    pass


class IncomeGeneratingActivityCategoryAddView(flourish.form.AddForm):

    template = InheritTemplate(flourish.page.Page.template)
    label = None
    legend = _('Category Information')
    fields = field.Fields(IIncomeGeneratingActivityCategory)

    def updateActions(self):
        super(IncomeGeneratingActivityCategoryAddView, self).updateActions()
        self.actions['add'].addClass('button-ok')
        self.actions['cancel'].addClass('button-cancel')

    @button.buttonAndHandler(STMessage('Submit'), name='add')
    def handleAdd(self, action):
        super(IncomeGeneratingActivityCategoryAddView, self).handleAdd.func(self, action)

    @button.buttonAndHandler(STMessage('Cancel'))
    def handleCancel(self, action):
        url = absoluteURL(self.context, self.request)
        self.request.response.redirect(url)

    def nextURL(self):
        return absoluteURL(self._category, self.request)

    def create(self, data):
        category = IncomeGeneratingActivityCategory()
        form.applyChanges(self, category, data)
        return category

    def add(self, category):
        chooser = INameChooser(self.context)
        name = chooser.chooseName(u'', category)
        self.context[name] = category
        self._category = category


class IncomeGeneratingActivityCategoryEditView(
        flourish.form.Form, form.EditForm):

    template = InheritTemplate(flourish.page.Page.template)
    content_template = ViewPageTemplateFile('templates/amount_form.pt')
    legend = _('Category Information')
    fields = field.Fields(IIncomeGeneratingActivityCategory)

    def update(self):
        form.EditForm.update(self)

    def updateActions(self):
        super(IncomeGeneratingActivityCategoryEditView, self).updateActions()
        self.actions['submit'].addClass('button-ok')
        self.actions['cancel'].addClass('button-cancel')

    def nextURL(self):
        return absoluteURL(self.context, self.request)

    @button.buttonAndHandler(STMessage('Submit'), name='submit')
    def handle_submit(self, action):
        super(IncomeGeneratingActivityCategoryEditView, self).handleApply.func(self, action)
        if (self.status == self.successMessage or
           self.status == self.noChangesMessage):
            self.request.response.redirect(self.nextURL())

    @button.buttonAndHandler(STMessage('Cancel'), name='cancel')
    def handle_cancel(self, action):
        self.request.response.redirect(self.nextURL())


class IncomeGeneratingActivityCategoryContainerAbsoluteURL(flourish.page.Page):

    adapts(IIncomeGeneratingActivityCategoryContainer,
           flourish.interfaces.IFlourishLayer)
    implements(IAbsoluteURL)

    def __str__(self):
        app = ISchoolToolApplication(None)
        url = str(getMultiAdapter((app, self.request), name='absolute_url'))
        return url + '/income_generating_activity_categories'

    __call__ = __str__


class IncomeGeneratingActivityCategoryView(flourish.form.DisplayForm):

    template = InheritTemplate(flourish.page.Page.template)
    content_template = ViewPageTemplateFile(
        'templates/income_activity_category.pt')

    @property
    def fields(self):
        result = field.Fields(IIncomeGeneratingActivityCategory)
        schema_field = Bool(__name__='default', title=_(u'Default'))
        result += field.Fields(schema_field)
        return result

    def getContent(self):
        container = self.context.__parent__
        return {
            'title': self.context.title,
            'default': sameProxiedObjects(self.context, container.default),
        }

    @property
    def canModify(self):
        return checkPermission('schooltool.edit', self.context)

    @property
    def title(self):
        return self.context.title

    def nextURL(self):
        return absoluteURL(self.context.__parent__, self.request)

    def edit_url(self):
        return '%s/edit.html' % absoluteURL(self.context, self.request)


class IncomeGeneratingActivityCategoryDoneLink(flourish.viewlet.Viewlet):

    template = flourish.templates.Inline('''
    <h3 class="done-link" i18n:domain="schooltool">
      <a tal:attributes="href view/view/nextURL"
         i18n:translate="">Done</a>
    </h3>
    ''')


class IncomeGeneratingActivityDoneLink(flourish.viewlet.Viewlet):

    template = flourish.templates.Inline('''
    <h3 class="done-link" i18n:domain="schooltool">
      <a tal:attributes="href view/view/nextURL"
         i18n:translate="">Done</a>
    </h3>
    ''')


class IncomeGeneratingActivityCategoryContainerDoneLink(
        flourish.viewlet.Viewlet):

    template = flourish.templates.Inline('''
    <h3 class="done-link" i18n:domain="schooltool">
      <a tal:attributes="href view/view/nextURL"
         i18n:translate="">Done</a>
    </h3>
    ''')


class IncomeGeneratingActivityCategoryActionsLinks(
        flourish.page.RefineLinksViewlet):

    pass


class IncomeGeneratingActivityCategoryMakeDefaultLink(
        flourish.page.LinkViewlet):

    @property
    def enabled(self):
        container = self.context.__parent__
        return not sameProxiedObjects(self.context, container.default)


class IncomeGeneratingActivityCategoryDeleteLink(flourish.page.LinkViewlet):

    @property
    def url(self):
        return '%s/delete.html?delete.%s&CONFIRM' % (
            absoluteURL(self.context.__parent__, self.request),
            self.context.__name__.encode('utf-8'))


class IncomeGeneratingActivityCategoryContainerDeleteView(
        flourish.containers.ContainerDeleteView):

    def nextURL(self):
        if 'CONFIRM' in self.request:
            return absoluteURL(self.context, self.request)
        return flourish.containers.ContainerDeleteView.nextURL(self)


class IncomeGeneratingActivityCategoryMakeDefaultView(flourish.page.Page):

    def __call__(self):
        container = removeSecurityProxy(self.context.__parent__)
        container.default = removeSecurityProxy(self.context)
        self.request.response.redirect(absoluteURL(self.context, self.request))


class IncomeGeneratingActivityCloseView(flourish.page.Page):

    def __call__(self):
        removeSecurityProxy(self.context).closed = True
        self.request.response.redirect(absoluteURL(self.context, self.request))


class IncomeGeneratingActivityReOpenView(flourish.page.Page):

    def __call__(self):
        removeSecurityProxy(self.context).closed = False
        self.request.response.redirect(absoluteURL(self.context, self.request))


class IncomeGeneratingActivityActionsLinks(flourish.page.RefineLinksViewlet):

    pass


class IncomeGeneratingActivityCloseLink(flourish.page.LinkViewlet):

    @property
    def enabled(self):
        return not self.context.closed


class IncomeGeneratingActivityReOpenLink(flourish.page.LinkViewlet):

    @property
    def enabled(self):
        return self.context.closed


class IncomeGeneratingActivityDeleteLink(flourish.page.ModalFormLinkViewlet):

    @property
    def enabled(self):
        return self.context.closed

    @property
    def dialog_title(self):
        title = _(u'Delete ${activity}',
                  mapping={'activity': self.context.title})
        return translate(title, context=self.request)


class IncomeGeneratingActivityDeleteView(flourish.form.DialogForm,
                                         form.EditForm):

    dialog_submit_actions = ('apply',)
    dialog_close_actions = ('cancel',)
    label = None

    @button.buttonAndHandler(STMessage("Delete"), name='apply')
    def handleDelete(self, action):
        url = '%s/delete.html?delete.%s&CONFIRM' % (
            absoluteURL(self.context.__parent__, self.request),
            self.context.__name__.encode('utf-8'))
        self.request.response.redirect(url)
        # We never have errors, so just close the dialog.
        self.ajax_settings['dialog'] = 'close'

    @button.buttonAndHandler(_("Cancel"))
    def handle_cancel_action(self, action):
        pass

    def updateActions(self):
        super(IncomeGeneratingActivityDeleteView, self).updateActions()
        self.actions['apply'].addClass('button-ok')
        self.actions['cancel'].addClass('button-cancel')


class IncomeGeneratingActivityContainerDeleteView(
        flourish.containers.ContainerDeleteView):

    def nextURL(self):
        if 'CONFIRM' in self.request:
            return absoluteURL(self.context, self.request)
        return flourish.containers.ContainerDeleteView.nextURL(self)


class IncomeGeneratingActivityEditView(flourish.form.Form, form.EditForm):

    template = InheritTemplate(flourish.page.Page.template)
    legend = _('Income Generating Activity Information')
    fields = field.Fields(IIncomeGeneratingActivity).select(*field_order)

    @property
    def title(self):
        return self.context.title

    def update(self):
        form.EditForm.update(self)

    def updateActions(self):
        super(IncomeGeneratingActivityEditView, self).updateActions()
        self.actions['submit'].addClass('button-ok')
        self.actions['cancel'].addClass('button-cancel')

    def nextURL(self):
        return absoluteURL(self.context, self.request)

    @button.buttonAndHandler(STMessage('Submit'), name='submit')
    def handle_submit(self, action):
        super(IncomeGeneratingActivityEditView, self).handleApply.func(self, action)
        if (self.status == self.successMessage or
           self.status == self.noChangesMessage):
            self.request.response.redirect(self.nextURL())

    @button.buttonAndHandler(STMessage('Cancel'), name='cancel')
    def handle_cancel(self, action):
        self.request.response.redirect(self.nextURL())


class RequestIncomeGeneratingActivitySummaryView(RequestRemoteReportDialog):

    report_builder = 'income_generating_activity_summary.pdf'


class IncomeGeneratingActivitySummaryPDFView(PersonFeeSummaryPDFView):

    name = _('Activity Summary')
    message_title = _('activity summary report')

    @property
    def base_filename(self):
        return '%s_activity_summary' % (self.context.__name__)


class IncomeGeneratingActivityCategoryContainerLink(flourish.page.LinkViewlet):

    @property
    def url(self):
        app = ISchoolToolApplication(None)
        container = IIncomeGeneratingActivityCategoryContainer(app)
        return absoluteURL(container, self.request)
